# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
Common errors for the aiocoap library
"""

import warnings
import abc
import errno
from typing import Optional

from .numbers import codes
from . import util


class Error(Exception):
    """
    Base exception for all exceptions that indicate a failed request
    """


class HelpfulError(Error):
    def __str__(self):
        """User presentable string. This should start with "Error:", or with
        "Something Error:", because the context will not show that this was an
        error."""
        return type(self).__name__

    def extra_help(self, hints={}) -> Optional[str]:
        """Information printed at aiocoap-client or similar occasions when the
        error message itself may be insufficient to point the user in the right
        direction

        The `hints` dictonary may be populated with context that the caller
        has; the implementation must tolerate their absence. Currently
        established keys:

        * original_uri (str): URI that was attemted to access
        * request (Message): Request that was assembled to be sent
        """
        return None


class RenderableError(Error, metaclass=abc.ABCMeta):
    """
    Exception that can meaningfully be represented in a CoAP response
    """

    @abc.abstractmethod
    def to_message(self):
        """Create a CoAP message that should be sent when this exception is
        rendered"""


class ResponseWrappingError(Error):
    """
    An exception that is raised due to an unsuccessful but received response.

    A better relationship with :mod:`.numbers.codes` should be worked out to do
    ``except UnsupportedMediaType`` (similar to the various ``OSError``
    subclasses).
    """

    def __init__(self, coapmessage):
        self.coapmessage = coapmessage

    def to_message(self):
        return self.coapmessage

    def __repr__(self):
        return "<%s: %s %r>" % (
            type(self).__name__,
            self.coapmessage.code,
            self.coapmessage.payload,
        )


class ConstructionRenderableError(RenderableError):
    """
    RenderableError that is constructed from class attributes :attr:`code` and
    :attr:`message` (where the can be overridden in the constructor).
    """

    def __init__(self, message=None):
        if message is not None:
            self.message = message

    def to_message(self):
        from .message import Message

        return Message(code=self.code, payload=self.message.encode("utf8"))

    code = codes.INTERNAL_SERVER_ERROR  #: Code assigned to messages built from it
    message = ""  #: Text sent in the built message's payload


# This block is code-generated to make the types available to static checkers.
# The __debug__ check below ensures that it stays up to date.
class BadRequest(ConstructionRenderableError):
    code = codes.BAD_REQUEST


class Unauthorized(ConstructionRenderableError):
    code = codes.UNAUTHORIZED


class BadOption(ConstructionRenderableError):
    code = codes.BAD_OPTION


class Forbidden(ConstructionRenderableError):
    code = codes.FORBIDDEN


class NotFound(ConstructionRenderableError):
    code = codes.NOT_FOUND


class MethodNotAllowed(ConstructionRenderableError):
    code = codes.METHOD_NOT_ALLOWED


class NotAcceptable(ConstructionRenderableError):
    code = codes.NOT_ACCEPTABLE


class RequestEntityIncomplete(ConstructionRenderableError):
    code = codes.REQUEST_ENTITY_INCOMPLETE


class Conflict(ConstructionRenderableError):
    code = codes.CONFLICT


class PreconditionFailed(ConstructionRenderableError):
    code = codes.PRECONDITION_FAILED


class RequestEntityTooLarge(ConstructionRenderableError):
    code = codes.REQUEST_ENTITY_TOO_LARGE


class UnsupportedContentFormat(ConstructionRenderableError):
    code = codes.UNSUPPORTED_CONTENT_FORMAT


class UnprocessableEntity(ConstructionRenderableError):
    code = codes.UNPROCESSABLE_ENTITY


class TooManyRequests(ConstructionRenderableError):
    code = codes.TOO_MANY_REQUESTS


class InternalServerError(ConstructionRenderableError):
    code = codes.INTERNAL_SERVER_ERROR


class NotImplemented(ConstructionRenderableError):
    code = codes.NOT_IMPLEMENTED


class BadGateway(ConstructionRenderableError):
    code = codes.BAD_GATEWAY


class ServiceUnavailable(ConstructionRenderableError):
    code = codes.SERVICE_UNAVAILABLE


class GatewayTimeout(ConstructionRenderableError):
    code = codes.GATEWAY_TIMEOUT


class ProxyingNotSupported(ConstructionRenderableError):
    code = codes.PROXYING_NOT_SUPPORTED


class HopLimitReached(ConstructionRenderableError):
    code = codes.HOP_LIMIT_REACHED


if __debug__:
    _missing_codes = False
    _full_code = ""
    for code in codes.Code:
        if code.is_successful() or not code.is_response():
            continue
        classname = "".join(w.title() for w in code.name.split("_"))
        _full_code += f"""
class {classname}(ConstructionRenderableError):
    code = codes.{code.name}"""
        if classname not in locals():
            warnings.warn(f"Missing exception type: f{classname}")
            _missing_codes = True
            continue
        if locals()[classname].code != code:
            warnings.warn(
                f"Mismatched code for {classname}: Should be {code}, is {locals()[classname].code}"
            )
            _missing_codes = True
            continue
    if _missing_codes:
        warnings.warn(
            "Generated exception list is out of sync, should be:\n" + _full_code
        )

# More detailed versions of code based errors


class NoResource(NotFound):
    """
    Raised when resource is not found.
    """

    message = "Error: Resource not found!"

    def __init__(self):
        warnings.warn(
            "NoResource is deprecated in favor of NotFound",
            DeprecationWarning,
            stacklevel=2,
        )


class UnallowedMethod(MethodNotAllowed):
    """
    Raised by a resource when request method is understood by the server
    but not allowed for that particular resource.
    """

    message = "Error: Method not allowed!"


class UnsupportedMethod(MethodNotAllowed):
    """
    Raised when request method is not understood by the server at all.
    """

    message = "Error: Method not recognized!"


class NetworkError(HelpfulError):
    """Base class for all "something went wrong with name resolution, sending
    or receiving packages".

    Errors of these kinds are raised towards client callers when things went
    wrong network-side, or at context creation. They are often raised from
    socket.gaierror or similar classes, but these are wrapped in order to make
    catching them possible independently of the underlying transport."""

    def __str__(self):
        return f"Network error: {type(self).__name__}"

    def extra_help(self, hints={}):
        if isinstance(self.__cause__, OSError):
            if self.__cause__.errno == errno.ECONNREFUSED:
                # seen trying to reach any used address with the port closed
                return "The remote host could be reached, but reported that the requested port is not open. Check whether a CoAP server is running at the address, or whether it is running on a different port."
            if self.__cause__.errno == errno.EHOSTUNREACH:
                # seen trying to reach any unused local address
                return "No way of contacting the remote host could be found. This could be because a host on the local network is offline or firewalled. Tools for debugging in the next step could be ping or traceroute."
            if self.__cause__.errno == errno.ENETUNREACH:
                # seen trying to reach an IPv6 host through an IP literal from a v4-only system, or trying to reach 2001:db8::1
                return "No way of contacting the remote network could be found. This may be due to lack of IPv6 connectivity, lack of a concrete route (eg. trying to reach a private use network which there is no route to). Tools for debugging in the next step could be ping or traceroute."
            if self.__cause__.errno == errno.EACCES:
                # seen trying to reach the broadcast address of a local network
                return "The operating system refused to send the request. For example, this can occur when attempting to send broadcast requests instead of multicast requests."


class NoRequestInterface(RuntimeError, ConstructionRenderableError, NetworkError):
    code = codes.PROXYING_NOT_SUPPORTED
    message = "Error: No CoAP transport available for this scheme on any request interface."  # or address, but we don't take transport-indication in full yet

    def extra_help(self, hints={}):
        return "More transports may be enabled by installing extra optional dependencies. Alternatively, consider using a CoAP-CoAP proxy."


class ResolutionError(NetworkError):
    """Resolving the host component of a URI to a usable transport address was
    not possible"""

    def __str__(self):
        return f"Name resolution error: {self.args[0]}"


class MessageError(NetworkError):
    """Received an error from the remote on the CoAP message level (typically a
    RST)"""


class RemoteServerShutdown(NetworkError):
    """The peer a request was sent to in a stateful connection closed the
    connection around the time the request was sent"""


class TimeoutError(NetworkError):
    """Base for all timeout-ish errors.

    Like NetworkError, receiving this alone does not indicate whether the
    request may have reached the server or not.
    """

    def extra_help(self, hints={}):
        return "Neither a response nor an error was received. This can have a wide range of causes, from the address being wrong to the server being stuck."


class ConRetransmitsExceeded(TimeoutError):
    """A transport that retransmits CON messages has failed to obtain a response
    within its retransmission timeout.

    When this is raised in a transport, requests failing with it may or may
    have been received by the server.
    """


class RequestTimedOut(TimeoutError):
    """
    Raised when request is timed out.

    This error is currently not produced by aiocoap; it is deprecated. Users
    can now catch error.TimeoutError, or newer more detailed subtypes
    introduced later.
    """


class WaitingForClientTimedOut(TimeoutError):
    """
    Raised when server expects some client action:

        - sending next PUT/POST request with block1 or block2 option
        - sending next GET request with block2 option

    but client does nothing.

    This error is currently not produced by aiocoap; it is deprecated. Users
    can now catch error.TimeoutError, or newer more detailed subtypes
    introduced later.
    """


class ResourceChanged(Error):
    """
    The requested resource was modified during the request and could therefore
    not be received in a consistent state.
    """


class UnexpectedBlock1Option(Error):
    """
    Raised when a server responds with block1 options that just don't match.
    """


class UnexpectedBlock2(Error):
    """
    Raised when a server responds with another block2 than expected.
    """


class MissingBlock2Option(Error):
    """
    Raised when response with Block2 option is expected
    (previous response had Block2 option with More flag set),
    but response without Block2 option is received.
    """


class NotObservable(Error):
    """
    The server did not accept the request to observe the resource.
    """


class ObservationCancelled(Error):
    """
    The server claimed that it will no longer sustain the observation.
    """


class UnparsableMessage(Error):
    """
    An incoming message does not look like CoAP.

    Note that this happens rarely -- the requirements are just two bit at the
    beginning of the message, and a minimum length.
    """


class LibraryShutdown(Error):
    """The library or a transport registered with it was requested to shut
    down; this error is raised in all outstanding requests."""


class AnonymousHost(Error):
    """This is raised when it is attempted to express as a reference a (base)
    URI of a host or a resource that can not be reached by any process other
    than this.

    Typically, this happens when trying to serialize a link to a resource that
    is hosted on a CoAP-over-TCP or -WebSockets client: Such resources can be
    accessed for as long as the connection is active, but can not be used any
    more once it is closed or even by another system."""


class MalformedUrlError(ValueError, HelpfulError):
    def __str__(self):
        if self.args:
            return f"Malformed URL: {self.args[0]}"
        else:
            return f"Malformed URL: {self.__cause__}"


class IncompleteUrlError(ValueError, HelpfulError):
    def __str__(self):
        return "URL incomplete: Must start with a scheme."

    def extra_help(self, hints={}):
        return "Most URLs in aiocoap need to be given with a scheme, eg. the 'coap' in 'coap://example.com/path'."


class MissingRemoteError(HelpfulError):
    """A request is sent without a .remote attribute"""

    def __str__(self):
        return "Error: No remote endpoint set for request."

    def extra_help(self, hints={}):
        original_uri = hints.get("original_uri", None)
        requested_message = hints.get("request", None)
        if requested_message and (
            requested_message.opt.proxy_uri or requested_message.opt.proxy_scheme
        ):
            if original_uri:
                return f"The message is set up for use with a proxy (because the scheme of {original_uri!r} is not supported), but no proxy was set."
            else:
                return (
                    "The message is set up for use with a proxy, but no proxy was set."
                )
        # Nothing helpful otherwise: The message was probably constructed in a
        # rather manual way.


__getattr__ = util.deprecation_getattr(
    {
        "UnsupportedMediaType": "UnsupportedContentFormat",
        "RequestTimedOut": "TimeoutError",
        "WaitingForClientTimedOut": "TimeoutError",
    },
    globals(),
)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tests for resource observation

Note that cancellation of observations is checked in neither way; that's
largely because the implementation has fallen behind on the drafts anyway and
needs to be updated."""

import asyncio
import aiocoap
from aiocoap.numbers import ContentFormat
import gc
import unittest

from aiocoap.resource import ObservableResource, WKCResource
from .test_server import (
    WithTestServer,
    WithClient,
    no_warnings,
    ReplacingResource,
    MultiRepresentationResource,
    asynctest,
)
from .common import run_fixture_as_standalone_server


class ObservableCounter(ObservableResource):
    def __init__(self, number_formatter=lambda x: x):
        super(ObservableCounter, self).__init__()
        self.count = 0
        self.number_formatter = number_formatter

    async def render_delete(self, request):
        self.count = 0
        self.updated_state()
        return aiocoap.Message(code=aiocoap.CHANGED)

    async def render_post(self, request):
        if request.payload == b"double":
            # Calling updated_state() twice without yielding inbetween is key
            # here; this ensures that Futures are not treated carelessly.
            self.count += 1
            # Triggering with explicit value because if left empty, the value
            # would be synthesized after the next yielding anyway, and no
            # ill-effects would be visible.
            self.updated_state(
                aiocoap.Message(
                    code=aiocoap.CONTENT,
                    payload=str(self.number_formatter(self.count)).encode("ascii"),
                )
            )
        self.count += 1
        self.updated_state()
        return aiocoap.Message(code=aiocoap.CHANGED)

    async def render_get(self, request):
        return aiocoap.Message(
            code=aiocoap.CONTENT,
            payload=str(self.number_formatter(self.count)).encode("ascii"),
        )

    async def render_fetch(self, request):
        return aiocoap.Message(
            code=aiocoap.CONTENT,
            payload=(
                "%s; request had length %s"
                % (self.number_formatter(self.count), len(request.payload))
            ).encode("ascii"),
        )


class ObservableReplacingResource(ReplacingResource, ObservableResource):
    async def render_put(self, request):
        result = await super(ObservableReplacingResource, self).render_put(request)

        self.updated_state()

        return result


class ObserveLateUnbloomer(ObservableResource):
    """A resource that accepts the server observation at first but at rendering
    time decides it can't do it"""

    def __init__(self):
        super().__init__()
        self._cancel_right_away = []

    async def add_observation(self, request, serverobservation):
        self._cancel_right_away.append(
            lambda: serverobservation.deregister("Changed my mind at render time")
        )
        serverobservation.accept(lambda: None)

    async def render_get(self, request):
        while self._cancel_right_away:
            self._cancel_right_away.pop(0)()
        return aiocoap.Message()


class ObservableFailure(ObservableResource):
    async def render_get(self, request):
        return aiocoap.Message(code=aiocoap.UNAUTHORIZED)


class ObserveTestingSite(aiocoap.resource.Site):
    prefix = []

    def __init__(self):
        super(ObserveTestingSite, self).__init__()

        self.add_resource(
            self.prefix + ["unobservable"],
            MultiRepresentationResource(
                {
                    ContentFormat.TEXT: b"",
                }
            ),
        )
        self.add_resource(self.prefix + ["count"], ObservableCounter())
        self.add_resource(self.prefix + ["echo"], ObservableReplacingResource())
        self.add_resource(self.prefix + ["notreally"], ObserveLateUnbloomer())
        self.add_resource(self.prefix + ["failure"], ObservableFailure())
        self.add_resource(
            self.prefix + ["large"], ObservableCounter(lambda x: (" %3d" % x) * 400)
        )


class NestedSite(aiocoap.resource.Site):
    def __init__(self):
        super().__init__()

        # Not part of the test suite, but handy when running standalone
        self.add_resource(
            [".well-known", "core"], WKCResource(self.get_resources_as_linkheader)
        )

        self.subsite = ObserveTestingSite()

        self.add_resource(["deep"], self.subsite)


class UnnestedSite(ObserveTestingSite):
    prefix = ["deep"]


class WithObserveTestServer(WithTestServer):
    def create_testing_site(self):
        self.testingsite = NestedSite()
        # use this when you suspect that things go wrong due to nesting;
        # usually not running this because it has no way to fail without nested
        # failing too
        # self.testingsite = UnnestedSite()
        return self.testingsite


class TestObserve(WithObserveTestServer, WithClient):
    @no_warnings
    @asynctest
    async def test_normal_get(self):
        request = aiocoap.Message(code=aiocoap.GET)
        request.opt.uri_path = ["deep", "count"]
        request.unresolved_remote = self.servernetloc

        response = await self.client.request(request).response
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Normal request did not succede"
        )
        self.assertEqual(
            response.payload, b"0", "Normal request gave unexpected result"
        )

    def build_observer(self, path, baserequest=None):
        if baserequest is not None:
            request = baserequest
        else:
            request = aiocoap.Message(code=aiocoap.GET)
        request.unresolved_remote = self.servernetloc
        request.opt.uri_path = path
        request.opt.observe = 0

        requester = self.client.request(request)
        observation_results = []

        async def task():
            try:
                async for obs in requester.observation:
                    observation_results.append(obs.payload)
                # This is how non-observability is indicated: The loop just terminates
                observation_results.append("done")
            except Exception as e:
                observation_results.append(e)

        running_task = asyncio.create_task(task())

        return requester, observation_results, running_task.cancel

    @no_warnings
    @asynctest
    async def test_unobservable(self):
        requester, observation_results, notinterested = self.build_observer(
            ["deep", "unobservable"]
        )

        response = await requester.response
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Unobservable base request did not succede"
        )
        self.assertEqual(
            response.payload, b"", "Unobservable base request gave unexpected result"
        )

        await asyncio.sleep(0.1)
        self.assertEqual(observation_results, ["done"])

    async def _change_counter(self, method, payload, path=["deep", "count"]):
        request = aiocoap.Message(code=method, uri_path=path, payload=payload)
        request.unresolved_remote = self.servernetloc
        await self.client.request(request).response_raising

    @asynctest
    async def _test_counter(
        self, baserequest, formatter, postpayload=b"", path=["deep", "count"]
    ):
        """Run a counter test with requests built from baserequest. Expect
        response payloads to be equal to the formatter(n) for n being the
        counter value"""
        await self._change_counter(aiocoap.DELETE, b"", path=path)

        requester, observation_results, notinterested = self.build_observer(
            path, baserequest=baserequest
        )

        response = await requester.response
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Observe base request did not succede"
        )
        self.assertEqual(
            response.payload,
            formatter(0),
            "Observe base request gave unexpected result",
        )

        await self._change_counter(aiocoap.POST, postpayload, path=path)
        await asyncio.sleep(0.1)
        self.assertEqual(observation_results, [formatter(1)])

        await self._change_counter(aiocoap.POST, postpayload, path=path)
        await asyncio.sleep(0.1)
        self.assertEqual(observation_results, [formatter(1), formatter(2)])

        notinterested()

    @no_warnings
    def test_counter(self):
        self._test_counter(None, lambda x: str(x).encode("ascii"))

    @no_warnings
    def test_counter_blockwise(self):
        self._test_counter(
            None,
            lambda x: str((" %3d" % x) * 400).encode("ascii"),
            path=["deep", "large"],
        )

    @no_warnings
    def test_counter_fetch(self):
        self._test_counter(
            aiocoap.Message(code=aiocoap.FETCH, payload=b"12345"),
            lambda x: ("%s; request had length 5" % x).encode("ascii"),
        )

    # Test hard disabled because not only it expects failure, but that failure
    # also causes protocol GC issues. Tracked in
    # https://github.com/chrysn/aiocoap/issues/95
    #     @no_warnings
    #     def test_counter_fetch_big(self):
    #         self._test_counter(
    #                 aiocoap.Message(code=aiocoap.FETCH, payload=b'12345' * 1000),
    #                 lambda x: ('%s; request had length 5000'%x).encode('ascii'))

    @no_warnings
    def test_counter_double(self):
        # see comments on b"double" in render_post
        self._test_counter(None, lambda x: str(x * 2).encode("ascii"), b"double")

    @no_warnings
    @asynctest
    async def test_echo(self):
        async def put(b):
            m = aiocoap.Message(code=aiocoap.PUT, payload=b)
            m.unresolved_remote = self.servernetloc
            m.opt.uri_path = ["deep", "echo"]
            response = await self.client.request(m).response
            self.assertEqual(response.code, aiocoap.CHANGED)

        await put(b"test data 1")

        requester, observation_results, notinterested = self.build_observer(
            ["deep", "echo"]
        )
        response = await requester.response
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Observe base request did not succede"
        )
        self.assertEqual(
            response.payload,
            b"test data 1",
            "Observe base request gave unexpected result",
        )

        await put(b"test data 2")

        await asyncio.sleep(0.1)
        self.assertEqual(observation_results, [b"test data 2"])

        notinterested()

    @unittest.expectedFailure  # regression since 82b35c1f8f, tracked as
    @no_warnings  # https://github.com/chrysn/aiocoap/issues/104
    @asynctest
    async def test_lingering(self):
        """Simulate what happens when a request is sent with an observe option,
        but the code only waits for the response and does not subscribe to the
        observation."""
        request = aiocoap.Message(code=aiocoap.GET)
        request.unresolved_remote = self.servernetloc
        request.opt.uri_path = ["deep", "count"]
        request.opt.observe = 0

        requester = self.client.request(request)

        response = await requester.response
        del requester, response
        gc.collect()

        # this needs to happen now and not in a precise_warnings because by the
        # time precise_warnings checks the messages, the context was already
        # shut down, but we want to make sure the warning is raised in time.
        self.assertWarned("Observation deleted without explicit cancellation")

    @no_warnings
    @asynctest
    async def test_unknownhost(self):
        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://cant.resolve.this.example./empty", observe=0
        )
        requester = self.client.request(request)

        events = []

        async def task():
            try:
                async for obs in requester.observation:
                    events.append("Callback: %s" % obs.payload)
                events.append("Regular empty return")
            except Exception as e:
                events.append("Error %s" % type(e).__name__)

        pull_task = asyncio.create_task(task())

        response = await requester.response_nonraising
        await asyncio.sleep(0.1)
        pull_task.cancel()

        self.assertEqual(events, ["Error ResolutionError"])

    @no_warnings
    @asynctest
    async def test_late_subscription_eventual_consistency(self):
        await self._change_counter(aiocoap.DELETE, b"")

        request = aiocoap.Message(code=aiocoap.GET)
        request.unresolved_remote = self.servernetloc
        request.opt.uri_path = ("deep", "count")
        request.opt.observe = 0

        requester = self.client.request(request)

        first_response = await requester.response
        self.assertEqual(
            first_response.payload, b"0", "Observe base request gave unexpected result"
        )

        await self._change_counter(aiocoap.POST, b"")
        await self._change_counter(aiocoap.POST, b"")

        last_seen = None

        async def task():
            nonlocal last_seen
            async for obs in requester.observation:
                last_seen = obs.payload

        pull_task = asyncio.create_task(task())

        # this is not required in the current implementation as it calls back
        # right from the registration, but i don't want to prescribe that.
        wait_a_moment = asyncio.get_running_loop().create_future()
        self.loop.call_soon(lambda: wait_a_moment.set_result(None))
        await wait_a_moment

        pull_task.cancel()

        self.assertEqual(last_seen, b"2")
        # only testing what was last seen because both receiving b"1" and b"2"
        # and only b"2" are correct eventually consistent results.

        requester.observation.cancel()

    async def _test_no_observe(self, path):
        m = aiocoap.Message(code=aiocoap.GET, observe=0)
        m.unresolved_remote = self.servernetloc
        m.opt.uri_path = path

        request = self.client.request(m)

        response = await request.response

        self.assertEqual(response.opt.observe, None)

        return request

    @no_warnings
    @asynctest
    async def test_notreally(self):
        await self._test_no_observe(["deep", "notreally"])

    @no_warnings
    @asynctest
    async def test_failure(self):
        request = await self._test_no_observe(["deep", "failure"])

        failure = None

        async def task():
            nonlocal failure
            try:
                async for obs in requester.observation:
                    pass
            except Exception as e:
                failure = e

        pull_task = asyncio.create_task(task())
        await asyncio.sleep(0.1)

        self.assertTrue(
            failure is not None, "Errback was not called on a failed observation"
        )


if __name__ == "__main__":
    # due to the imports, you'll need to run this as `python3 -m tests.test_observe`
    run_fixture_as_standalone_server(WithObserveTestServer)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Confront a CoAP over TCP server with a client that speaks so bad protocol it
is easier to mock with sending byte sequences than with aiocoap"""

import asyncio
import unittest
import aiocoap

from .fixtures import precise_warnings, no_warnings, asynctest
from .test_server import WithTestServer
from .common import tcp_disabled


@unittest.skipIf(tcp_disabled, "TCP disabled in environment")
class TestNoncoapTCPClient(WithTestServer):
    def setUp(self):
        super().setUp()

        self.mock_r, self.mock_w = self.loop.run_until_complete(
            asyncio.open_connection(self.serveraddress, aiocoap.COAP_PORT)
        )

    def tearDown(self):
        self.mock_w.close()

        super().tearDown()

    @staticmethod
    def _read_as_messages(encoded: bytes):
        """Process the encoded data into CoAP-over-TCP messages, return them as
        a list and trailing (unrecognized / incomplete) data."""
        messages = []
        while True:
            size = aiocoap.transports.tcp._extract_message_size(encoded)
            if size is not None:
                size = sum(size)
            if size is None or size > len(encoded):
                return messages, encoded

            messages.append(aiocoap.transports.tcp._decode_message(encoded[:size]))
            encoded = encoded[size:]

    async def should_abort_early(self, request: bytes):
        """Send request bytes, expect that the server closes the connection
        after having sent possibly a CSM and an abort"""
        self.mock_w.write(request)
        r = (
            await self.mock_r.read()
        )  # timing out would be a typical failure case here too
        parsed, trail = self._read_as_messages(r)
        self.assertEqual(trail, b"", "Leftover data after closing message")
        if parsed[0].code == aiocoap.CSM:
            # don't discard the CSM unconditionallly: the server might have
            # read the request data before sending its own initial CSM.
            parsed.pop(0)
        self.assertEqual(
            len(parsed), 1, "Not exactly one (presumably abort) message received"
        )
        self.assertEqual(
            parsed[0].code, aiocoap.ABORT, "Received message is not an abort message"
        )

    async def should_idle(self, request: bytes, timeout=0.1):
        """Send request bytes, expect that the server sends CSM and does not
        close the connection, awaiting more from the client.

        Returns all messages received until the timeout."""
        self.mock_w.write(request)
        triggered_eof = False

        async def kill_read():
            """After a timeout, synthesize an end-of-file condition into the
            reader, hoping this doesn't beak too much."""
            nonlocal triggered_eof
            await asyncio.sleep(timeout)
            triggered_eof = True
            self.mock_r.feed_eof()

        self.loop.create_task(kill_read())
        r = (
            await self.mock_r.read()
        )  # timing out would be a typical failure case here too
        self.assertEqual(triggered_eof, True, "Server closed connection prematurely")

        parsed, trail = self._read_as_messages(r)
        # if this happens, the server is either sending garbage (announcing
        # something long and not following up), or the timeout should be
        # increased
        self.assertEqual(trail, b"", "Leftover data after reading timeout")
        if parsed[0].code == aiocoap.CSM:
            # don't discard the CSM unconditionallly: the server might have
            # read the request data before sending its own initial CSM.
            parsed.pop(0)

        return parsed

    async def should_idle_quietly(self, request: bytes, timeout=0.1):
        """should_idle, but assert that no messages were returned"""

        messages = await self.should_idle(request, timeout)

        # it's not a per-spec wrong thing to do, but highly unusual
        self.assertEqual(messages, [], "Server sent messages on its own")

    @precise_warnings(["Aborting connection: Failed to parse message"])
    @asynctest
    async def test_http_get(self):
        await self.should_abort_early(b"GET /.well-known/core HTTP/1.0")

    @precise_warnings(["Aborting connection: No CSM received"])
    @asynctest
    async def test_early_get(self):
        await self.should_abort_early(b"\0\x01")

    @no_warnings
    @asynctest
    async def test_incomplete_small(self):
        await self.should_idle_quietly(b"\0")

    @no_warnings
    @asynctest
    async def test_incomplete_large1(self):
        # announcing but not sending 1 bytes extlen
        await self.should_idle_quietly(b"\xd0")

    @no_warnings
    @asynctest
    async def test_incomplete_large2(self):
        # sending one out of four bytes extlen
        # a server could in theory reject this on grounds of "no matter what
        # you say next, my buffer ain't large enough"
        await self.should_idle_quietly(b"\xf0\0")

    @no_warnings
    @asynctest
    async def test_incomplete_large3(self):
        # announcing a 269 byte long message, but not even sendin the code
        await self.should_idle_quietly(b"\xe0\0\0")

    @precise_warnings(["Aborting connection: Overly large message announced"])
    @asynctest
    async def test_incomplete_large4(self):
        # announcing the longest possible message, this should excede
        # everyone's max-message-size.
        #
        # blocking to read more would be acceptable behavior as well.
        await self.should_abort_early(b"\xf0\xff\xff\xff\xff")

    @precise_warnings(["Aborting connection: Failed to parse message"])
    @asynctest
    async def test_wrong_tkl(self):
        # send an unspecified token length of 15.
        # the rest of the message is an empty CSM, so if the server were to
        # extrapolate from the meaning of tkl 0..8, it'd read it as OK.
        await self.should_abort_early(b"\x0fxxxxxxxxxxxxxxx\xe1")

    # Fun inside the CSM

    @no_warnings
    @asynctest
    async def test_exotic_elective_csm_option(self):
        # send option number something-even (something-odd plus 269) as an empty option
        await self.should_idle_quietly(b"\x30\xe1\xe0\xf1\xf1")

    @precise_warnings(["Aborting connection: Option not supported"])
    @asynctest
    async def test_exotic_compulsory_csm_option(self):
        # send option number something-odd (something-even plus 269) as an empty option
        await self.should_abort_early(b"\x30\xe1\xe0\xf2\xf2")

    @precise_warnings(["Aborting connection: Option not supported"])
    @asynctest
    async def test_exotic_compulsory_csm_option_late(self):
        # send an empty CSM, and after that the one from compulsory_csm_option
        await self.should_abort_early(b"\0\xe1\x30\xe1\xe0\xf2\xf2")



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This test runs a resource directory and executes the examples from the RD
specification against it.
"""

import unittest

import aiocoap
from aiocoap.util import hostportjoin

from .test_server import WithAsyncLoop, Destructing, WithClient, asynctest

linkheader_modules = aiocoap.defaults.linkheader_missing_modules()
_skip_unless_linkheader = unittest.skipIf(
    linkheader_modules,
    "Modules missing for running RD tests: %s" % (linkheader_modules,),
)
if not linkheader_modules:
    from aiocoap.util.linkformat import link_header
    import aiocoap.cli.rd


class WithResourceDirectory(WithAsyncLoop, Destructing):
    rd_address = "::1"
    rd_port = 56830
    rd_netloc = "[%s]:%d" % (rd_address, rd_port)

    def setUp(self):
        super().setUp()

        self.rd = aiocoap.cli.rd.Main(
            ["--bind", hostportjoin("::1", self.rd_port)], loop=self.loop
        )
        self.loop.run_until_complete(self.rd.initializing)

    def tearDown(self):
        self.loop.run_until_complete(self.rd.shutdown())
        super().tearDown()

        self._del_to_be_sure("rd")


class TestDiscovery(WithResourceDirectory, WithClient):
    @_skip_unless_linkheader
    @asynctest
    async def test_discovery(self):
        request = aiocoap.Message(
            code=aiocoap.GET,
            uri="coap://%s/.well-known/core?rt=core.rd*" % self.rd_netloc,
        )
        response = await self.client.request(request).response

        self.assertEqual(
            response.code, aiocoap.CONTENT, "RD discovery did not give content"
        )
        links = link_header.parse(response.payload.decode("utf8"))
        # Not checking for presence of group resources: not implemented here
        for rt in ("core.rd", "core.rd-lookup-ep", "core.rd-lookup-res"):
            self.assertEqual(
                len([x for x in links.links if x.rt == [rt]]),
                1,
                "Not exactly one entry of rt=%s found" % rt,
            )

    async def _get_endpoint(self, rt):
        """Return the URI for a given rt in the configured RD"""

        if not hasattr(self, "_endpoints"):
            request = aiocoap.Message(
                code=aiocoap.GET,
                uri="coap://%s/.well-known/core?rt=core.rd*" % self.rd_netloc,
            )
            response = await self.client.request(request).response

            self._endpoints = {
                entry.rt[0]: entry.get_target(response.get_request_uri())
                for entry in link_header.parse(response.payload.decode("utf8")).links
            }

        return self._endpoints[rt]

    @_skip_unless_linkheader
    @asynctest
    async def test_registration(self):
        request = aiocoap.Message(
            code=aiocoap.POST,
            uri=(await self._get_endpoint("core.rd")) + "?ep=node1",
            content_format=40,
            payload=b'</sensors/temp>;ct=41;rt="temperature-c";if="sensor",</sensors/light>;ct=41;rt="light-lux";if="sensor"',
        )
        response = await self.client.request(request).response

        self.assertEqual(
            response.code, aiocoap.CREATED, "Registration did not result in Created"
        )
        self.assertTrue(
            len(response.opt.location_path) > 0,
            "Registration did not result in non-empty registration resource",
        )

    # FIXME: there are many more things to be tested here



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import json
import tempfile
import shutil
import subprocess
import unittest
import sys

import aiocoap

from .test_server import WithClient, WithTestServer

from .fixtures import no_warnings, asynctest
from .common import tcp_disabled, run_fixture_as_standalone_server

IS_STANDALONE = False


class WithTLSServer(WithTestServer):
    def setUp(self):
        self.keydir = tempfile.mkdtemp(suffix="-testkeypair")
        self.keyfile = self.keydir + "/key.pem"
        self.certfile = self.keydir + "/cert.pem"
        self.credentialsfile = self.keydir + "/credentials.json"
        subprocess.check_call(
            [
                "openssl",
                "req",
                "-x509",
                "-newkey",
                "rsa:4096",
                "-keyout",
                self.keyfile,
                "-out",
                self.certfile,
                "-days",
                "5",
                "-nodes",
                "-subj",
                "/CN=%s" % self.servernamealias,
            ],
            stderr=subprocess.DEVNULL,
        )

        # Write out for the benefit of standalone clients during debugging
        with open(self.credentialsfile, "w") as of:
            json.dump(
                {
                    "coaps+tcp://%s/*" % self.servernamealias: {
                        "tlscert": {"certfile": self.certfile}
                    }
                },
                of,
            )

        if IS_STANDALONE:
            print(
                "To test, run ./aiocoap-client coaps+tcp://%s/whoami --credentials %s"
                % (
                    self.servernamealias,
                    self.credentialsfile,
                )
            )

        super().setUp()

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.keydir)

    def get_server_ssl_context(self):
        import ssl

        # FIXME: copied from aiocoap.cli.common
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
        ssl_context.set_alpn_protocols(["coap"])
        ssl_context.sni_callback = lambda obj, name, context: setattr(
            obj, "indicated_server_name", name
        )
        return ssl_context


class WithTLSClient(WithClient):
    # This expects that something -- typically the colocated WithTestServer -- sets certfile first
    def setUp(self):
        super().setUp()

        self.client.client_credentials["coaps+tcp://%s/*" % self.servernamealias] = (
            aiocoap.credentials.TLSCert(certfile=self.certfile)
        )


@unittest.skipIf(tcp_disabled, "TCP disabled in environment")
class TestTLS(WithTLSServer, WithTLSClient):
    @no_warnings
    @asynctest
    async def test_tls(self):
        request = aiocoap.Message(code=aiocoap.GET)
        request.set_request_uri(
            "coaps+tcp://%s/whoami" % self.servernamealias, set_uri_host=False
        )
        response = await self.client.request(request).response_raising

        response = json.loads(response.payload)
        self.assertEqual(
            response["requested_uri"],
            "coaps+tcp://%s/whoami" % self.servernamealias,
            "SNI name was not used by the server",
        )


if __name__ == "__main__":
    # due to the imports, you'll need to run this as `python3 -m tests.test_server`
    IS_STANDALONE = True
    import logging

    logging.basicConfig(level=logging.DEBUG)
    run_fixture_as_standalone_server(TestTLS)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio

from aiocoap.util.asyncio.timeoutdict import TimeoutDict

from .fixtures import WithAsyncLoop, asynctest


class TestTimeoutDict(WithAsyncLoop):
    @asynctest
    async def test_presence_and_absence(self):
        timeout = 0.2
        d = TimeoutDict(timeout)
        d["k"] = "v"
        await asyncio.sleep(timeout / 2)

        assert d["k"] == "v"

        await asyncio.sleep(timeout * 2.5)
        self.assertRaises(KeyError, lambda: d["k"])



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import copy

import aiocoap
import aiocoap.optiontypes

import unittest


class TestMessage(unittest.TestCase):
    def test_encode(self):
        msg1 = aiocoap.Message(mtype=aiocoap.CON, mid=0, code=aiocoap.EMPTY)
        binary1 = bytes((64, 0, 0, 0))
        self.assertEqual(
            msg1.encode(), binary1, "wrong encode operation for empty CON message"
        )

        msg2 = aiocoap.Message(
            mtype=aiocoap.ACK,
            mid=0xBC90,
            code=aiocoap.CONTENT,
            payload=b"temp = 22.5 C",
            token=b"q",
        )
        msg2.opt.etag = b"abcd"
        binary2 = (
            bytes((97, 69, 188, 144, 113, 68))
            + b"abcd"
            + bytes((255,))
            + b"temp = 22.5 C"
        )
        self.assertEqual(
            msg2.encode(),
            binary2,
            "wrong encode operation for ACK message with payload, and Etag option",
        )
        msg2short = binary2[0:5] + binary2[10:]  # header, token, marker, data
        msg2a = copy.deepcopy(msg2)
        del msg2a.opt.etag
        self.assertEqual(
            msg2a.encode(), msg2short, "deleting single property did not succeed"
        )
        msg2b = copy.deepcopy(msg2)
        del msg2b.opt.etags
        self.assertEqual(
            msg2b.encode(), msg2short, "deleting list property did not succeed"
        )
        msg2c = copy.deepcopy(msg2)
        msg2c.opt.etags = []
        self.assertEqual(
            msg2c.encode(), msg2short, "emptying list property did not succeed"
        )
        msg2d = copy.deepcopy(msg2)
        msg2d.opt.etag = None
        self.assertEqual(
            msg2d.encode(), msg2short, "setting single property to None did not succeed"
        )

        msg3 = aiocoap.Message()
        self.assertRaises(TypeError, msg3.encode)

        msg4 = aiocoap.Message(mtype=aiocoap.CON, mid=2 << 16)
        self.assertRaises(Exception, msg4.encode)

        msg5 = aiocoap.Message(mtype=aiocoap.CON, mid=0, code=aiocoap.EMPTY)
        o = aiocoap.optiontypes.OpaqueOption(1234, value=b"abcd")
        msg5.opt.add_option(o)
        binary5 = binary1 + bytes((0xE4, 0x03, 0xC5)) + b"abcd"
        self.assertEqual(
            msg5.encode(), binary5, "wrong encoding for high option numbers"
        )

        msg6 = aiocoap.Message(mtype=aiocoap.CON, mid=0, code=aiocoap.EMPTY)
        o = aiocoap.optiontypes.OpaqueOption(12345678, value=b"abcd")
        msg6.opt.add_option(o)
        self.assertRaises(ValueError, msg6.encode)

        msg7 = aiocoap.Message(mtype=aiocoap.CON, mid=0, code=aiocoap.EMPTY)

        def set_unknown_opt():
            msg7.opt.foobar = 42

        self.assertRaises(AttributeError, set_unknown_opt)

    def test_decode(self):
        rawdata1 = bytes((64, 0, 0, 0))
        self.assertEqual(
            aiocoap.Message.decode(rawdata1).mtype,
            aiocoap.CON,
            "wrong message type for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata1).mid,
            0,
            "wrong message ID for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata1).code,
            aiocoap.EMPTY,
            "wrong message code for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata1).token,
            b"",
            "wrong message token for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata1).payload,
            b"",
            "wrong message payload for decode operation",
        )
        rawdata2 = (
            bytes((97, 69, 188, 144, 113, 68))
            + b"abcd"
            + bytes((255,))
            + b"temp = 22.5 C"
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).mtype,
            aiocoap.ACK,
            "wrong message type for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).mid,
            0xBC90,
            "wrong message ID for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).code,
            aiocoap.CONTENT,
            "wrong message code for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).token,
            b"q",
            "wrong message token for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).payload,
            b"temp = 22.5 C",
            "wrong message payload for decode operation",
        )
        self.assertEqual(
            aiocoap.Message.decode(rawdata2).opt.etags,
            (b"abcd",),
            "problem with etag option decoding for decode operation",
        )
        self.assertEqual(
            len(aiocoap.Message.decode(rawdata2).opt._options),
            1,
            "wrong number of options after decode operation",
        )
        rawdata3 = rawdata1 + bytes((0xF0,))
        self.assertRaises(
            aiocoap.error.UnparsableMessage, aiocoap.Message.decode, rawdata3
        )  # message with option delta reserved for payload marker
        rawdata4 = rawdata1 + bytes((0xE0,))
        self.assertRaises(
            aiocoap.error.UnparsableMessage, aiocoap.Message.decode, rawdata4
        )  # message with extended option delta that is not actually there
        rawdata5 = rawdata1 + bytes((0x04,))
        self.assertRaises(
            aiocoap.error.UnparsableMessage, aiocoap.Message.decode, rawdata5
        )  # message with option that is not actually there


class TestReadExtendedFieldValue(unittest.TestCase):
    def test_read_extended_field_value(self):
        arguments = (
            (0, b"aaaa"),
            (0, b""),
            (1, b"aaaa"),
            (12, b"aaaa"),
            (13, b"aaaa"),
            (13, b"a"),
            (14, b"aaaa"),
            (14, b"aa"),
        )
        results = (
            (0, b"aaaa"),
            (0, b""),
            (1, b"aaaa"),
            (12, b"aaaa"),
            (110, b"aaa"),
            (110, b""),
            (25198, b"aa"),
            (25198, b""),
        )

        for argument, result in zip(arguments, results):
            self.assertEqual(
                aiocoap.options._read_extended_field_value(argument[0], argument[1]),
                result,
                "wrong result for value : "
                + repr(argument[0])
                + " , rawdata : "
                + repr(argument[1]),
            )


class TestUintOption(unittest.TestCase):
    def test_encode(self):
        arguments = (0, 1, 2, 40, 50, 255, 256, 1000)
        results = (
            b"",
            bytes((1,)),
            bytes((2,)),
            bytes((40,)),
            bytes((50,)),
            bytes((255,)),
            bytes((1, 0)),
            bytes((3, 232)),
        )
        for argument, result in zip(arguments, results):
            self.assertEqual(
                aiocoap.optiontypes.UintOption(0, argument).encode(),
                result,
                "wrong encode operation for option value : " + str(argument),
            )

    def test_decode(self):
        arguments = (
            b"",
            bytes((1,)),
            bytes((2,)),
            bytes((40,)),
            bytes((50,)),
            bytes((255,)),
            bytes((1, 0)),
            bytes((3, 232)),
        )
        results = (0, 1, 2, 40, 50, 255, 256, 1000)
        for argument, result in zip(arguments, results):
            o = aiocoap.optiontypes.UintOption(0)
            o.decode(argument)
            self.assertEqual(
                o.value, result, "wrong decode operation for rawdata : " + str(argument)
            )


class TestOptions(unittest.TestCase):
    def test_set_uri_path(self):
        opt1 = aiocoap.options.Options()
        opt1.uri_path = ["core"]
        self.assertEqual(
            len(opt1.get_option(aiocoap.OptionNumber.URI_PATH)),
            1,
            "wrong uri_path setter operation for single string argument",
        )
        self.assertEqual(
            opt1.get_option(aiocoap.OptionNumber.URI_PATH)[0].value,
            "core",
            "wrong uri_path setter operation for single string argument",
        )
        opt2 = aiocoap.options.Options()
        opt2.uri_path = ("core", ".well-known")
        self.assertEqual(
            len(opt2.get_option(aiocoap.OptionNumber.URI_PATH)),
            2,
            "wrong uri_path setter operation for 2-element tuple argument",
        )
        self.assertEqual(
            opt2.get_option(aiocoap.OptionNumber.URI_PATH)[0].value,
            "core",
            "wrong uri_path setter operation for 2-element tuple argument",
        )
        self.assertEqual(
            opt2.get_option(aiocoap.OptionNumber.URI_PATH)[1].value,
            ".well-known",
            "wrong uri_path setter operation for 2-element tuple argument",
        )
        opt3 = aiocoap.options.Options()
        self.assertRaises(TypeError, setattr, opt3, "uri_path", 42)


class TestOptiontypes(unittest.TestCase):
    def test_optiontypes(self):
        # from rfc725 table 4
        on = aiocoap.numbers.OptionNumber
        options = {
            on.IF_MATCH: "C",
            on.URI_HOST: "CU",
            on.ETAG: "",
            on.MAX_AGE: "U",
            on.SIZE1: "N",
        }

        for o, expected in options.items():
            self.assertEqual(
                "C" in expected, o.is_critical(), "Unexpected criticalness of %r" % o
            )
            self.assertEqual(
                "C" not in expected,
                o.is_elective(),
                "Unexpected electiveness of %r" % o,
            )
            self.assertEqual(
                "U" in expected, o.is_unsafe(), "Unexpected unsafeness of %r" % o
            )
            self.assertEqual(
                "U" not in expected,
                o.is_safetoforward(),
                "Unexpected safetoforwardness of %r" % o,
            )
            if o.is_safetoforward():
                self.assertEqual(
                    "N" in expected,
                    o.is_nocachekey(),
                    "Unexpected nocachekeyness of %r" % o,
                )
                self.assertEqual(
                    "N" not in expected,
                    o.is_cachekey(),
                    "Unexpected cachekeyness of %r" % o,
                )
            else:
                self.assertRaises(ValueError, o.is_nocachekey)
                self.assertRaises(ValueError, o.is_cachekey)


class TestMessageOptionConstruction(unittest.TestCase):
    def test_uri_construction(self):
        message = aiocoap.Message(uri="coap://some-host:1234/some/path/")
        self.assertEqual(message.opt.uri_host, "some-host")
        self.assertEqual(message.opt.uri_port, None)
        self.assertEqual(message.opt.uri_path, ("some", "path", ""))

    def test_opt_roundtripping(self):
        message = aiocoap.Message()
        message.opt.content_format = 40
        message.opt.etag = b"1234"
        message.opt.uri_path = ()
        self.assertEqual(message.opt.content_format, 40)
        self.assertEqual(message.opt.etag, b"1234")
        self.assertEqual(message.opt.uri_path, ())

    def test_opt_construction(self):
        message = aiocoap.Message(content_format=40, observe=b"1234", uri_path=())
        self.assertEqual(message.opt.content_format, 40)
        self.assertEqual(message.opt.observe, b"1234")
        self.assertEqual(message.opt.uri_path, ())

    def test_copy(self):
        message = aiocoap.Message()
        original_state = repr(message)
        new_one = message.copy(
            payload=b"x",
            mid=42,
            code=0,
            mtype=3,
            token=b"xyz",
            observe=0,
            content_format=1234,
        )
        self.assertEqual(original_state, repr(message), "Message.copy mutated original")
        self.assertEqual(new_one.payload, b"x")
        self.assertEqual(new_one.mid, 42)
        self.assertEqual(new_one.token, b"xyz")
        self.assertEqual(str(new_one.code), "EMPTY")
        # "Type.RST" is also ok if the enum decides to have a different str,
        # but it should be of mtype class and not just a number. Beware that
        # other tests also look for the string versions (eg. "ACK EMPTY")
        self.assertEqual(str(new_one.mtype), "RST")
        self.assertEqual(new_one.opt.observe, 0)
        self.assertEqual(new_one.opt.content_format, 1234)

        new_two = new_one.copy(uri="coap://some-host/some/path")
        self.assertEqual(new_two.opt.uri_path, ("some", "path"))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import aiocoap

import unittest


class TestMessage(unittest.TestCase):
    normal_uris = [
        "coap://hostname:1234/path",
        "coap://hostname:1234/path?query=string&argument=x",
        "coap+tcp://hostname/path",
        "coaps://hostname:1234/path",
    ]

    def test_normal(self):
        for n in self.normal_uris:
            msg = aiocoap.Message(code=aiocoap.GET, uri=n)
            self.assertEqual(
                msg.get_request_uri(),
                n,
                "Encoding normal URI to a message did not round-trip the URI",
            )

    # These do not map URIs to their fully normalzied form, just to another
    # (possibly normalized) form that is equivalent under normalization rules.
    # Thus, it might be perfectly OK for an implemention change to require
    # changing the test case -- as long as the new result sill has the same
    # normal URI.
    denormal_uris = {
        "coap://hostname:1234": "coap://hostname:1234/",
        "CoAp://HoStNaMe/": "coap://hostname/",
        "coap://host/%7Esensors": "coap://host/~sensors",
        "coap://hostname:5683/path": "coap://hostname:5683/path",  # could strip the port
        "coap://host/blåbærsyltetøy": "coap://host/bl%C3%A5b%C3%A6rsyltet%C3%B8y",
    }

    def test_denormal(self):
        for src, dest in self.denormal_uris.items():
            msg = aiocoap.Message(code=aiocoap.GET, uri=src)
            self.assertEqual(
                msg.get_request_uri(),
                dest,
                "Encoding normal URI to a message did not"
                " round-trip the URI (original: %r)" % src,
            )

    erring_uris = {
        "/hello": aiocoap.error.IncompleteUrlError,
        "coap://[": aiocoap.error.MalformedUrlError,
        "coap://example.com/%ff": aiocoap.error.MalformedUrlError,
        "coap://example.com:fivesixeightthree/": aiocoap.error.MalformedUrlError,
        # not broken per URI specs, but CoAP uses the hostname component,
        # and urllib uses information about which schemes use it.
        "coap:like:urn": aiocoap.error.MalformedUrlError,
    }

    def test_errors(self):
        for uri, expected in self.erring_uris.items():
            with self.assertRaises(expected):
                aiocoap.Message(code=aiocoap.GET, uri=uri)

    uris_using_proxy = {
        "http://example.com/test",
        "urn:uuid:6e8bc430-9c3a-11d9-9669-0800200c9a66",
    }

    def test_turned_into_proxy(self):
        for uri in self.uris_using_proxy:
            msg = aiocoap.Message(code=aiocoap.GET, uri=uri)
            self.assertEqual(msg.opt.proxy_uri, uri)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import unittest

from .test_server import WithClient, WithTestServer, TestServer
import secrets
import aiocoap.defaults

oscore_modules = aiocoap.defaults.oscore_missing_modules()

if not oscore_modules:
    import aiocoap.oscore
    import aiocoap.oscore_sitewrapper
    import cbor2

_skip_unless_oscore = unittest.skipIf(
    oscore_modules, "Modules missing for running OSCORE tests: %s" % (oscore_modules,)
)


class WithGroupKeys(unittest.TestCase):
    def _set_up_algorithms(self):
        self.alg_aead = aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM]
        self.alg_group_enc = aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM]
        self.hashfun = aiocoap.oscore.hashfunctions[aiocoap.oscore.DEFAULT_HASHFUNCTION]
        self.alg_countersign = aiocoap.oscore.Ed25519()
        self.alg_pairwise_key_agreement = aiocoap.oscore.EcdhSsHkdf256()

    def setUp(self):
        self._set_up_algorithms()

        group_id = b"G"
        participants = [b"", b"\x01", b"longname"]
        private_keys, creds = zip(
            *(self.alg_countersign.generate_with_ccs() for _ in participants)
        )
        master_secret = secrets.token_bytes(64)
        master_salt = b"PoCl4"
        # This would only be processed when there is actual contact with the GM
        gm_cred = b"dummy credential"

        self.groups = [
            aiocoap.oscore.SimpleGroupContext(
                self.alg_aead,
                self.hashfun,
                self.alg_countersign,
                self.alg_group_enc,
                self.alg_pairwise_key_agreement,
                group_id,
                master_secret,
                master_salt,
                participants[i],
                private_keys[i],
                creds[i],
                {
                    participants[j]: creds[j]
                    for j, _ in enumerate(participants)
                    if i != j
                },
                gm_cred,
                group_manager_cred_fmt="dummy",
            )
            for i, _ in enumerate(participants)
        ]

        super().setUp()


class WithGroupServer(WithTestServer, WithGroupKeys):
    def setUp(self):
        super().setUp()

        server_credentials = aiocoap.credentials.CredentialsMap()
        server_credentials[":a"] = self.groups[0]
        self.server.serversite = aiocoap.oscore_sitewrapper.OscoreSiteWrapper(
            self.server.serversite, server_credentials
        )


class WithGroupClient(WithClient):
    def setUp(self):
        super().setUp()

        self.client.client_credentials["coap://%s/*" % self.servernetloc] = self.groups[
            1
        ]


@_skip_unless_oscore
class TestGroupOscore(TestServer, WithGroupServer, WithGroupClient):
    pass


@_skip_unless_oscore
class TestGroupOscoreWithPairwise(TestGroupOscore):
    def setUp(self):
        super().setUp()

        for k, v in self.client.client_credentials.items():
            self.client.client_credentials[k] = v.pairwise_for(self.groups[0].sender_id)


@_skip_unless_oscore
class TestDifferentLengths(TestGroupOscore):
    def _set_up_algorithms(self):
        super()._set_up_algorithms()
        # Override Group Encryption Algorithm to have 16 bytes nonce rather than the 13 of alg_aead
        self.alg_group_enc = aiocoap.oscore.A128CBC


# For the different length tests, it completely suffices to just run one.
for _testname in dir(TestDifferentLengths):
    if not _testname.startswith("test_"):
        continue
    if _testname != "test_big_resource":
        setattr(TestDifferentLengths, _testname, None)
assert any(
    getattr(TestDifferentLengths, _t) is not None
    for _t in dir(TestDifferentLengths)
    if _t.startswith("test_")
), "Removing tests left none"


@_skip_unless_oscore
class TestDifferentLengthsWithPairwise(
    TestDifferentLengths, TestGroupOscoreWithPairwise
):
    pass


del TestServer



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Implementations of RFC9176_

Currently, this only contains the :mod:`client` module. There is an
implementation of the server side contained in aiocoap, but it is currently
accessible only as :mod:`command line application <aiocoap.cli.rd>`.

.. _RFC9176: https://datatracker.ietf.org/doc/html/rfc9176
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""List of known values for the CoAP "Code" field.

The values in this module correspond to the IANA registry "`CoRE Parameters`_",
subregistries "CoAP Method Codes" and "CoAP Response Codes".

The codes come with methods that can be used to get their rough meaning, see
the :class:`Code` class for details.

.. _`CoRE Parameters`: https://www.iana.org/assignments/core-parameters/core-parameters.xhtml
"""

import warnings

from ..util import ExtensibleIntEnum


class Code(ExtensibleIntEnum):
    """Value for the CoAP "Code" field.

    As the number range for the code values is separated, the rough meaning of
    a code can be determined using the :meth:`is_request`, :meth:`is_response` and
    :meth:`is_successful` methods."""

    EMPTY = 0
    GET = 1
    POST = 2
    PUT = 3
    DELETE = 4
    FETCH = 5
    PATCH = 6
    iPATCH = 7
    CREATED = 65
    DELETED = 66
    VALID = 67
    CHANGED = 68
    CONTENT = 69
    CONTINUE = 95
    BAD_REQUEST = 128
    UNAUTHORIZED = 129
    BAD_OPTION = 130
    FORBIDDEN = 131
    NOT_FOUND = 132
    METHOD_NOT_ALLOWED = 133
    NOT_ACCEPTABLE = 134
    REQUEST_ENTITY_INCOMPLETE = 136
    CONFLICT = (4 << 5) + 9
    PRECONDITION_FAILED = 140
    REQUEST_ENTITY_TOO_LARGE = 141
    UNSUPPORTED_CONTENT_FORMAT = 143

    @property
    def UNSUPPORTED_MEDIA_TYPE(self):
        warnings.warn(
            "UNSUPPORTED_MEDIA_TYPE is a deprecated alias for UNSUPPORTED_CONTENT_FORMAT"
        )
        return self.UNSUPPORTED_CONTENT_FORMAT

    UNPROCESSABLE_ENTITY = (4 << 5) + 22
    TOO_MANY_REQUESTS = (4 << 5) + 29
    INTERNAL_SERVER_ERROR = 160
    NOT_IMPLEMENTED = 161
    BAD_GATEWAY = 162
    SERVICE_UNAVAILABLE = 163
    GATEWAY_TIMEOUT = 164
    PROXYING_NOT_SUPPORTED = 165
    HOP_LIMIT_REACHED = (5 << 5) + 8

    CSM = 225
    PING = 226
    PONG = 227
    RELEASE = 228
    ABORT = 229

    def is_request(self):
        """True if the code is in the request code range"""
        return True if (self >= 1 and self < 32) else False

    def is_response(self):
        """True if the code is in the response code range"""
        return True if (self >= 64 and self < 192) else False

    def is_signalling(self):
        return True if self >= 224 else False

    def is_successful(self):
        """True if the code is in the successful subrange of the response code range"""
        return True if (self >= 64 and self < 96) else False

    def can_have_payload(self):
        """True if a message with that code can carry a payload. This is not
        checked for strictly, but used as an indicator."""
        return self.is_response() or self in (
            self.POST,
            self.PUT,
            self.FETCH,
            self.PATCH,
            self.iPATCH,
        )

    @property
    def class_(self):
        """The class of a code (distinguishing whether it's successful, a
        request or a response error or more).

        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.CONTENT.class_
        2
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code.BAD_GATEWAY.class_
        5
        """
        return self >> 5

    @property
    def dotted(self):
        """The numeric value three-decimal-digits (c.dd) form"""
        return "%d.%02d" % divmod(self, 32)

    @property
    def name_printable(self):
        """The name of the code in human-readable form"""
        return self.name.replace("_", " ").title()

    def __str__(self):
        """
        >>> print(Code.GET)
        GET
        >>> print(Code.CONTENT)
        2.05 Content
        >>> print(Code.BAD_GATEWAY)
        5.02 Bad Gateway
        >>> print(Code(32))
        32
        """
        if self.is_request() or self is self.EMPTY:
            return self.name
        elif self.is_response() or self.is_signalling():
            return "%s %s" % (self.dotted, self.name_printable)
        else:
            return "%d" % self

    def _classification(self):
        return ("Successful " if self.is_successful() else "") + (
            "Request "
            if self.is_request()
            else "Response "
            if self.is_response()
            else ""
        )

    def __repr__(self):
        """
        >>> Code.GET
        <Request Code 1 "GET">
        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code(32)
        <Code 32 "32">
        """
        return '<%sCode %d "%s">' % (self._classification(), self, self)

    def _repr_html_(self):
        """
        >>> Code.GET._repr_html_()
        '<abbr title="Request Code 0.01">GET</abbr>'
        >>> Code(31)._repr_html_()
        '<abbr title="Unknown Request Code">0.31</abbr>'
        """
        import html

        if self.name == "(unknown)":
            return f'<abbr title="Unknown {self._classification()}Code">{self.dotted}</abbr>'
        else:
            return f'<abbr title="{self._classification()}Code {self.dotted}">{html.escape(self.name)}</abbr>'

    @classmethod
    def _missing_(cls, value):
        constructed = super()._missing_(value)
        constructed._name_ = "(unknown)"
        return constructed


# List is copied down to help mypy and other typing dependent tools to
# understand the types.
EMPTY = Code.EMPTY
GET = Code.GET
POST = Code.POST
PUT = Code.PUT
DELETE = Code.DELETE
FETCH = Code.FETCH
PATCH = Code.PATCH
iPATCH = Code.iPATCH
CREATED = Code.CREATED
DELETED = Code.DELETED
VALID = Code.VALID
CHANGED = Code.CHANGED
CONTENT = Code.CONTENT
CONTINUE = Code.CONTINUE
BAD_REQUEST = Code.BAD_REQUEST
UNAUTHORIZED = Code.UNAUTHORIZED
BAD_OPTION = Code.BAD_OPTION
FORBIDDEN = Code.FORBIDDEN
NOT_FOUND = Code.NOT_FOUND
METHOD_NOT_ALLOWED = Code.METHOD_NOT_ALLOWED
NOT_ACCEPTABLE = Code.NOT_ACCEPTABLE
REQUEST_ENTITY_INCOMPLETE = Code.REQUEST_ENTITY_INCOMPLETE
CONFLICT = Code.CONFLICT
PRECONDITION_FAILED = Code.PRECONDITION_FAILED
REQUEST_ENTITY_TOO_LARGE = Code.REQUEST_ENTITY_TOO_LARGE
UNSUPPORTED_CONTENT_FORMAT = Code.UNSUPPORTED_CONTENT_FORMAT
UNPROCESSABLE_ENTITY = Code.UNPROCESSABLE_ENTITY
TOO_MANY_REQUESTS = Code.TOO_MANY_REQUESTS
INTERNAL_SERVER_ERROR = Code.INTERNAL_SERVER_ERROR
NOT_IMPLEMENTED = Code.NOT_IMPLEMENTED
BAD_GATEWAY = Code.BAD_GATEWAY
SERVICE_UNAVAILABLE = Code.SERVICE_UNAVAILABLE
GATEWAY_TIMEOUT = Code.GATEWAY_TIMEOUT
PROXYING_NOT_SUPPORTED = Code.PROXYING_NOT_SUPPORTED
HOP_LIMIT_REACHED = Code.HOP_LIMIT_REACHED
CSM = Code.CSM
PING = Code.PING
PONG = Code.PONG
RELEASE = Code.RELEASE
ABORT = Code.ABORT
if __debug__:
    for _code in Code:
        if locals().get(_code.name) is not _code:
            warnings.warn(
                f"Locals list is out of sync; entry `{_code.name} = Code.{_code.name}` is missing"
            )


__all__ = ["Code"] + [
    k for (k, v) in locals().items() if isinstance(v, Code) and not k.startswith("_")
]



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import contextlib
import re
import aiocoap
import aiocoap.resource
from aiocoap.numbers import ContentFormat
import unittest
import os
import json

from . import common
from .fixtures import no_warnings, WithAsyncLoop, Destructing, CLEANUPTIME, asynctest


class slow_empty_ack(contextlib.ContextDecorator):
    def __enter__(self):
        # FIXME: Pushing the times here so that even on a loaded system
        # with pypy and coverage, this can complete realistically.
        #
        # This will break as soon as transport tuning becomes more
        # per-transport or even per-remote; the proper fix would be to use
        # mock time anyway.
        self.original_empty_ack_delay = aiocoap.numbers.TransportTuning.EMPTY_ACK_DELAY
        aiocoap.numbers.TransportTuning.EMPTY_ACK_DELAY = (
            0.9 * aiocoap.numbers.TransportTuning.ACK_TIMEOUT
        )
        return self

    def __exit__(self, *exc):
        aiocoap.numbers.TransportTuning.EMPTY_ACK_DELAY = self.original_empty_ack_delay
        return False


class MultiRepresentationResource(aiocoap.resource.Resource):
    def __init__(self, representations):
        self._representations = representations
        super().__init__()

    async def render_get(self, request):
        if request.opt.proxy_scheme is not None:
            # For CLI test_noproxy -- but otherwise a symptom of https://github.com/chrysn/aiocoap/issues/268
            return aiocoap.Message(code=aiocoap.CONTENT, payload=b"This is no proxy")

        m = request.opt.accept or ContentFormat.TEXT

        if m in self._representations:
            response = self._representations[m]
        else:
            return aiocoap.Message(code=aiocoap.NOT_ACCEPTABLE)

        return aiocoap.Message(payload=response, content_format=request.opt.accept or 0)


class SlowResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        await asyncio.sleep(0.2)
        # Marking the response as confirmable overrides the default behavior of
        # sending NON responses to NON requests.
        #
        # This is done to aid the test_freeoncancel test, and should revert to
        # the default behavior once that has better control over the environment.
        return aiocoap.Message(mtype=aiocoap.CON)


class BigResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        # 10kb
        payload = b"0123456789----------" * 512
        response = aiocoap.Message(payload=payload)

        aiocoap.resource.hashing_etag(request, response)
        return response


class SlowBigResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        # Should be 0.2s usually, but while running in slow_empty_ack mode, we
        # have to slow it down. Adds 2s to the tests, but that makes the test
        # more reliable.
        await asyncio.sleep(aiocoap.numbers.TransportTuning.EMPTY_ACK_DELAY * 1.1)
        # 1.6kb
        payload = b"0123456789----------" * 80
        return aiocoap.Message(payload=payload)


class ManualBigResource(aiocoap.resource.Resource):
    async def needs_blockwise_assembly(self, request):
        return False

    async def render_get(self, request):
        BlockwiseTuple = aiocoap.optiontypes.BlockOption.BlockwiseTuple
        block2 = request.opt.block2 or BlockwiseTuple(0, 0, 6)
        # as above
        body = b"0123456789----------" * 80
        # in a more realistic example, we wouldn't build this in memory but eg.
        # seek and read limited length
        slice = body[block2.start :]
        more = len(slice) > block2.size
        slice = slice[: block2.size]
        block2 = BlockwiseTuple(block2.block_number, more, block2.size_exponent)
        return aiocoap.Message(payload=slice, block2=block2)


class ReplacingResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        return aiocoap.Message(payload=self.value)

    async def render_put(self, request):
        self.value = request.payload.replace(b"0", b"O")
        return aiocoap.Message()

    async def render_post(self, request):
        response = request.payload.replace(b"0", b"O")
        return aiocoap.Message(code=aiocoap.CONTENT, payload=response)


class RootResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        return aiocoap.Message(
            code=aiocoap.CONTENT, payload=b"Welcome to the test server"
        )


class GenericErrorResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        raise RuntimeError()


class PrettyErrorResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        raise self.MyError()

    class MyError(aiocoap.error.ConstructionRenderableError):
        code = aiocoap.NOT_FOUND
        message = "I'm sorry nothing is here"


class DoubleErrorResource(aiocoap.resource.Resource):
    async def render_get(self, request):
        raise self.MyError()

    class MyError(aiocoap.error.RenderableError):
        def to_message(self):
            raise RuntimeError()


class WhoAmI(aiocoap.resource.Resource):
    async def render_get(self, request):
        p = dict(
            repr=repr(request.remote),
            hostinfo=request.remote.hostinfo,
            hostinfo_local=request.remote.hostinfo_local,
            scheme=request.remote.scheme,
            urihost_option=request.opt.uri_host,
            # FIXME: The whole get_request_uri is a mess
            requested_uri=request._original_request_uri,
            claims=request.remote.authenticated_claims,
        )
        return aiocoap.Message(
            code=aiocoap.CONTENT,
            content_format=ContentFormat.JSON,
            payload=json.dumps(p).encode("utf8"),
        )


class BasicTestingSite(aiocoap.resource.Site):
    def __init__(self):
        super(BasicTestingSite, self).__init__()

        # Not part of the test suite, but handy when running standalone
        self.add_resource(
            [".well-known", "core"],
            aiocoap.resource.WKCResource(self.get_resources_as_linkheader),
        )

        self.add_resource(
            ["empty"],
            MultiRepresentationResource(
                {
                    ContentFormat.JSON: b"{}",
                    ContentFormat.LINKFORMAT: b"<>",
                    ContentFormat.TEXT: b"",
                }
            ),
        )
        self.add_resource(
            ["answer"],
            MultiRepresentationResource(
                {
                    ContentFormat.JSON: b'{"answer": 42}',
                    ContentFormat.CBOR: b"\xa1\x66\x61\x6e\x73\x77\x65\x72\x18\x2a",
                    ContentFormat.LINKFORMAT: b'<data:text/plain;42>;rel="answer";anchor="https://en.wikipedia.org/wiki/Phrases_from_The_Hitchhiker%27s_Guide_to_the_Galaxy#Answer_to_the_Ultimate_Question_of_Life,_the_Universe,_and_Everything_(42)"',
                    ContentFormat.TEXT: b"The answer to life, the universe, and everything is 42.",
                }
            ),
        )
        self.add_resource(["slow"], SlowResource())
        self.add_resource(["big"], BigResource())
        self.add_resource(["slowbig"], SlowBigResource())
        self.add_resource(["manualbig"], ManualBigResource())
        self.add_resource(["replacing"], self.Subsite())
        self.add_resource(["error", "generic"], GenericErrorResource())
        self.add_resource(["error", "pretty"], PrettyErrorResource())
        self.add_resource(["error", "double"], DoubleErrorResource())
        self.add_resource(["whoami"], WhoAmI())
        self.add_resource([], RootResource())

    class Subsite(aiocoap.resource.Site):
        def __init__(self):
            super().__init__()
            self.add_resource(["one"], ReplacingResource())


class WithTestServer(WithAsyncLoop, Destructing):
    # to allow overriding the factory class
    TestingSite = BasicTestingSite

    def create_testing_site(self):
        return self.TestingSite()

    def get_server_ssl_context(self):
        """Override hook for subclasses that want to populate _ssl_context at construction"""
        return None

    def setUp(self):
        super(WithTestServer, self).setUp()

        multicastif = (
            os.environ["AIOCOAP_TEST_MCIF"].split(":")
            if "AIOCOAP_TEST_MCIF" in os.environ
            else []
        )

        self.server = self.loop.run_until_complete(
            aiocoap.Context.create_server_context(
                self.create_testing_site(),
                bind=(self.serveraddress, None),
                multicast=multicastif,
                _ssl_context=self.get_server_ssl_context(),
            )
        )

    def tearDown(self):
        # let the server receive the acks we just sent
        self.loop.run_until_complete(asyncio.sleep(CLEANUPTIME))
        self.loop.run_until_complete(self.server.shutdown())
        # Nothing in the context should keep the request interfaces alive;
        # delete them first to see *which* of them is the one causing the
        # trouble
        while self.server.request_interfaces:
            self._del_to_be_sure(
                {
                    "get": (lambda self: self.server.request_interfaces[0]),
                    "del": (lambda self: self.server.request_interfaces.__delitem__(0)),
                    "label": "request_interfaces[%s]"
                    % self.server.request_interfaces[0],
                }
            )
        self._del_to_be_sure("server")

        super(WithTestServer, self).tearDown()

    serveraddress = "::1"
    servernetloc = "[%s]" % serveraddress
    servernamealias = common.loopbackname_v6 or common.loopbackname_v46


class WithClient(WithAsyncLoop, Destructing):
    def setUp(self):
        super(WithClient, self).setUp()

        self.test_did_shut_down_client = False

        self.client = self.loop.run_until_complete(
            aiocoap.Context.create_client_context()
        )

    def tearDown(self):
        if not self.test_did_shut_down_client:
            self.test_did_shut_down_client = False
            self.loop.run_until_complete(asyncio.sleep(CLEANUPTIME))
            self.loop.run_until_complete(self.client.shutdown())

        # Nothing in the context should keep the request interfaces alive;
        # delete them first to see *which* of them is the one causing the
        # trouble
        while self.client.request_interfaces:
            self._del_to_be_sure(
                {
                    "get": (lambda self: self.client.request_interfaces[0]),
                    "del": (lambda self: self.client.request_interfaces.__delitem__(0)),
                    "label": "request_interfaces[%s]"
                    % self.client.request_interfaces[0],
                }
            )

        self._del_to_be_sure("client")

        super(WithClient, self).tearDown()


# test cases


class TestServerBase(WithTestServer, WithClient):
    """All the tools for building requests, but no actual tests; use this when
    working off the test server with no intention to to the full set of
    tests."""

    @no_warnings
    def build_request(self):
        request = aiocoap.Message(code=aiocoap.GET)
        request.unresolved_remote = self.servernetloc
        return request

    @asynctest
    async def fetch_response(self, request):
        return await self.client.request(request).response


class TestServer(TestServerBase):
    @no_warnings
    def test_empty_accept(self):
        request = self.build_request()
        request.opt.uri_path = ["empty"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Simple request did not succede"
        )
        self.assertEqual(response.payload, b"", "Simple request gave unexpected result")

    @no_warnings
    def test_unacceptable_accept(self):
        request = self.build_request()
        request.opt.uri_path = ["empty"]
        request.opt.accept = 9999
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.NOT_ACCEPTABLE,
            "Inacceptable request was not not accepted",
        )

    @no_warnings
    def test_js_accept(self):
        request = self.build_request()
        request.opt.uri_path = ["empty"]
        request.opt.accept = ContentFormat.JSON
        response = self.fetch_response(request)
        self.assertEqual(response.code, aiocoap.CONTENT, "JSON request did not succede")
        self.assertEqual(response.payload, b"{}", "JSON request gave unexpected result")

    @no_warnings
    def test_nonexisting_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["nonexisting"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.NOT_FOUND, "Nonexisting resource was not not found"
        )

    @no_warnings
    def test_spurious_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["..", "empty"]
        response = self.fetch_response(request)
        # different behavior would be ok-ish, as the .. in the request is forbidden, but returning 4.04 is sane here
        self.assertEqual(
            response.code,
            aiocoap.NOT_FOUND,
            "'..' component in path did not get ignored the way it was expected",
        )

    @no_warnings
    @slow_empty_ack()
    def test_fast_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["empty"]

        response = self.fetch_response(request)

        self.assertEqual(response.code, aiocoap.CONTENT, "Fast request did not succede")
        self.assertEqual(self._count_empty_acks(), 0, "Fast resource had an empty ack")

    @no_warnings
    def test_slow_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["slow"]

        response = self.fetch_response(request)

        self.assertEqual(response.code, aiocoap.CONTENT, "Slow request did not succede")
        if response.requested_scheme in (None, "coap"):
            self.assertEqual(
                self._count_empty_acks(),
                1,
                "Slow resource was not handled in two exchanges",
            )

    @no_warnings
    def test_big_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["big"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.CONTENT, "Big resource request did not succede"
        )
        self.assertEqual(
            len(response.payload), 10240, "Big resource is not as big as expected"
        )

        self.assertTrue(
            response.opt.etag != None, "Big resource does not provide an ETag"
        )

        request = self.build_request()
        request.opt.uri_path = ["big"]
        request.opt.etags = [response.opt.etag]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.VALID,
            "Big resource does not support ETag validation",
        )
        self.assertTrue(
            response.opt.etag != None, "Big resource does not send ETag for validation"
        )

    @no_warnings
    @slow_empty_ack()
    def test_slowbig_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["slowbig"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.CONTENT, "SlowBig resource request did not succede"
        )
        self.assertEqual(
            len(response.payload), 1600, "SlowBig resource is not as big as expected"
        )
        if response.requested_scheme in (None, "coap"):
            self.assertEqual(
                self._count_empty_acks(),
                1,
                "SlowBig resource was not handled in two exchanges",
            )

    @no_warnings
    def test_manualbig_resource(self):
        request = self.build_request()
        request.opt.uri_path = ["manualbig"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.CONTENT, "ManualBig resource request did not succede"
        )
        self.assertEqual(
            len(response.payload), 1600, "ManualBig resource is not as big as expected"
        )

    @no_warnings
    def test_replacing_resource(self):
        testpattern = b"01" * 1024

        request = self.build_request()
        request.code = aiocoap.PUT
        request.payload = testpattern
        request.opt.uri_path = ["replacing", "one"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code, aiocoap.CHANGED, "PUT did not result in CHANGED"
        )
        self.assertEqual(response.payload, b"", "PUT has unexpected payload")

        request = self.build_request()
        request.code = aiocoap.GET
        request.opt.uri_path = ["replacing", "one"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Replacing resource could not be GOT (GET'd?) successfully",
        )
        self.assertEqual(
            response.payload,
            testpattern.replace(b"0", b"O"),
            "Replacing resource did not replace as expected between PUT and GET",
        )

        request = self.build_request()
        request.code = aiocoap.POST
        request.payload = testpattern
        request.opt.uri_path = ["replacing", "one"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Replacing resource could not be POSTed to successfully",
        )
        self.assertEqual(
            response.payload,
            testpattern.replace(b"0", b"O"),
            "Replacing resource did not replace as expected when POSTed",
        )

    def test_error_resources(self):
        request = self.build_request()
        request.opt.uri_path = ["error", "generic"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.INTERNAL_SERVER_ERROR,
            "Runtime error possibly leaking information",
        )
        self.assertEqual(
            response.payload, b"", "Runtime error possibly leaking information"
        )

        request = self.build_request()
        request.opt.uri_path = ["error", "pretty"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.NOT_FOUND,
            "Runtime error possibly leaking information",
        )
        self.assertTrue(
            response.payload.decode("ascii").startswith("I'm sorry"),
            "Runtime error possibly leaking information",
        )

        request = self.build_request()
        request.opt.uri_path = ["error", "double"]
        response = self.fetch_response(request)
        self.assertEqual(
            response.code,
            aiocoap.INTERNAL_SERVER_ERROR,
            "Runtime double error possibly leaking information",
        )
        self.assertEqual(
            response.payload, b"", "Runtime double error possibly leaking information"
        )

    @no_warnings
    def test_root_resource(self):
        request = self.build_request()
        request.opt.uri_path = []
        response = self.fetch_response(request)
        self.assertEqual(response.code, aiocoap.CONTENT, "Root resource was not found")

    _empty_ack_logmsg = re.compile(
        "^Incoming message <aiocoap.Message at 0x[0-9a-f]+: ACK EMPTY ([^)]+)"
    )

    def _count_empty_acks(self):
        # only client-side received empty-acks are counted; they typically
        # generate an empty ack back when the separate response is ack'd.
        return sum(
            self._empty_ack_logmsg.match(x.msg) is not None
            for x in self.handler
            if x.name != "coap-server"
        )

    @no_warnings
    def test_clean_shutdown(self):
        self.loop.run_until_complete(self.client.shutdown())
        self.test_did_shut_down_client = True

        request = self.build_request()
        # This is clearly misuse of the API, but nonetheless, only a
        # LibraryShutdown should fly out of it.
        #
        # Ideally, we'd do that "during" shutdown", but that's terribly racy
        # business, both wins are acceptable (what if the shutdown hasn't
        # really begun yet), and it's highly unlikely that we hit the juicy
        # cases.
        with self.assertRaises(aiocoap.error.LibraryShutdown):
            self.loop.run_until_complete(self.fetch_response(request))


@unittest.skipIf(common.tcp_disabled, "TCP disabled in environment")
class TestServerTCP(TestServer):
    # no modification in server setup necessary, as by default, all transports
    # are enabled on servers.

    def build_request(self):
        request = super().build_request()
        request.requested_scheme = "coap+tcp"
        return request


ws_modules = aiocoap.defaults.ws_missing_modules()


@unittest.skipIf(
    ws_modules or common.ws_disabled,
    "WS missing modules (%s) or disabled in this environment" % (ws_modules,),
)
class TestServerWS(TestServer):
    # as with TestServerWS

    def build_request(self):
        request = super().build_request()
        # odd default port
        request.unresolved_remote += ":8683"
        request.requested_scheme = "coap+ws"
        return request


if __name__ == "__main__":
    # due to the imports, you'll need to run this as `python3 -m tests.test_server`
    common.run_fixture_as_standalone_server(WithTestServer)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import doctest
import aiocoap.defaults
import os
import sys


def _load_tests():
    """This is an adapted load_test of the old unittest mechanism, adapted
    crudely to work with pytest. It can most probably be done better, eg. by
    producing a list of doctest-able files for use with doctest-glob (which
    would allow getting actual breakpoints where things fail, or other pytest
    niceties), but as it is, it at least covers the cases again."""
    i = 0
    for root, dn, fn in os.walk("aiocoap"):
        for f in fn:
            if not f.endswith(".py"):
                continue
            p = os.path.join(root, f)
            p = p[:-3].replace(os.sep, ".")
            if (
                "oscore" in p or "edhoc" in p
            ) and aiocoap.defaults.oscore_missing_modules():
                continue
            if (
                "cryptography_additions" in p
                and aiocoap.defaults.oscore_missing_modules()
            ):
                continue
            if p.endswith(".ws") and aiocoap.defaults.ws_missing_modules():
                continue
            if (
                "resourcedirectory" in p
                or "fileserver" in p
                or p in ("aiocoap.cli.rd", "aiocoap.util.linkformat")
                and aiocoap.defaults.linkheader_missing_modules()
            ):
                continue
            if (
                p in ("aiocoap.util.prettyprint", "aiocoap.util.linkformat_pygments")
                and aiocoap.defaults.prettyprint_missing_modules()
            ):
                continue
            if (
                p in ("aiocoap.util.pyodide_websockets",)
                and not aiocoap.defaults.is_pyodide
            ):
                continue
            for t in doctest.DocTestSuite(p):
                i += 1

                def test(t=t):
                    result = t.run()
                    for f in result.failures:
                        print(f[1])
                        raise RuntimeError("Doctest failed (see above)")
                    for e in result.errors:
                        raise e

                globals()["test_%03d" % i] = test


_load_tests()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from pygments import token, lexers
from pygments.lexer import RegexLexer, bygroups

__all__ = ["LinkFormatLexer"]


class LinkFormatLexer(RegexLexer):
    name = "LinkFormatLexer"
    mimetypes = ["application/link-format"]

    tokens = {
        "root": [
            (
                "(<)([^>]*)(>)",
                bygroups(token.Punctuation, token.Name.Label, token.Punctuation),
                "maybe-end",
            )
        ],
        "maybe-end": [
            # Whitespace is not actually allowed, but produced by the pretty printer
            (";\\s*", token.Punctuation, "attribute"),
            (",\\s*", token.Punctuation, "root"),
        ],
        "attribute": [
            (
                '([^,;=]+)((=)("[^"]*"|[^,;"]+))?',
                bygroups(
                    token.Name.Attribute, None, token.Operator, token.String.Symbol
                ),
                "maybe-end",
            ),
        ],
    }


def _register():
    if "LinkFormatLexer" not in lexers.LEXERS:
        lexers.LEXERS["LinkFormatLexer"] = (
            "aiocoap.util.linkformat_pygments",
            "LinkFormatLexer",
            (),
            (),
            LinkFormatLexer.mimetypes,
        )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Internal module containing types used inside EDHOC security contexts"""

import abc
import enum
import io
from pathlib import Path
import random
from typing import Optional, Dict, Literal
import os

import cbor2
import lakers

from . import oscore, credentials, error
from . import Message
from .numbers import POST


def load_cbor_or_edn(filename: Path):
    """Common heuristic for whether something is CBOR or EDN"""
    import cbor_diag
    import cbor2

    with filename.open("rb") as binary:
        try:
            result = cbor2.load(binary)
        except cbor2.CBORDecodeError:
            pass
        else:
            if binary.read(1) == b"":
                return result
            # else it apparently hasn't been CBOR all through...
    with filename.open() as textual:
        try:
            converted = cbor_diag.diag2cbor(textual.read())
        except ValueError:
            raise credentials.CredentialsLoadError(
                "Data loaded from %s was recognized neither as CBOR nor CBOR Diagnostic Notation (EDN)"
                % filename
            )
        # no need to check for completeness: diag2cbor doesn't do diagnostic
        # sequences, AIU that's not even a thing
        return cbor2.loads(converted)


class CoseKeyForEdhoc:
    kty: int
    crv: int
    d: bytes

    @classmethod
    def from_file(cls, filename: Path) -> "CoseKeyForEdhoc":
        """Load a key from a file (in CBOR or EDN), asserting that the file is not group/world readable"""
        if filename.stat().st_mode & 0o077 != 0:
            raise credentials.CredentialsLoadError(
                "Refusing to load private key that is group or world accessible"
            )

        loaded = load_cbor_or_edn(filename)
        return cls.from_map(loaded)

    @classmethod
    def from_map(cls, key: dict) -> "CoseKeyForEdhoc":
        if not isinstance(key, dict):
            raise credentials.CredentialsLoadError(
                "Data is not shaped like COSE_KEY (expected top-level dictionary)"
            )
        if 1 not in key:
            raise credentials.CredentialsLoadError(
                "Data is not shaped like COSE_KEY (expected key 1 (kty) in top-level dictionary)"
            )
        if key[1] != 2:
            raise credentials.CredentialsLoadError(
                "Private key type %s is not supported (currently only 2 (EC) is supported)"
                % (key[1],)
            )

        if key.get(-1) != 1:
            raise credentials.CredentialsLoadError(
                "Private key of type EC requires key -1 (crv), currently supported values: 1 (P-256)"
            )

        if not isinstance(key.get(-4), bytes) or len(key[-4]) != 32:
            raise credentials.CredentialsLoadError(
                "Private key of type EC P-256 requires key -4 (d) to be a 32-byte long byte string"
            )

        if any(k not in (1, -1, -4) for k in key):
            raise credentials.CredentialsLoadError(
                "Extraneous data in key, consider allow-listing the item if acceptable"
            )

        s = cls()
        s.kty = 2  # EC
        s.crv = 1  # P-256
        s.d = key[-4]

        return s

    def secret_to_map(self) -> dict:
        # kty: EC, crv: P-256, d: ...
        return {1: self.kty, -1: self.crv, -4: self.d}

    # Should we deprecate filename, add a generate_in_file method? (It's there
    # because generate originally depended on a file system)
    @classmethod
    def generate(cls, filename: Optional[Path] = None) -> "CoseKeyForEdhoc":
        """Generate a key inside a file

        This returns the generated private key.
        """

        from cryptography.hazmat.primitives.asymmetric import ec

        key = ec.generate_private_key(curve=ec.SECP256R1())

        s = cls()
        s.kty = 2  # EC
        s.crv = 1  # P-256
        s.d = key.private_numbers().private_value.to_bytes(32, "big")

        if filename is not None:
            flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
            if hasattr(os, "O_BINARY"):
                flags |= os.O_BINARY
            descriptor = os.open(filename, flags, mode=0o600)
            try:
                with open(descriptor, "wb") as keyfile:
                    cbor2.dump(s.secret_to_map(), keyfile)
            except Exception:
                filename.unlink()
                raise

        return s

    def as_ccs(
        self, kid: Optional[bytes], subject: Optional[str]
    ) -> Dict[Literal[14], dict]:
        """Given a key, generate a corresponding KCCS"""

        from cryptography.hazmat.primitives.asymmetric import ec

        private = ec.derive_private_key(int.from_bytes(self.d, "big"), ec.SECP256R1())
        public = private.public_key()

        x = public.public_numbers().x.to_bytes(32, "big")
        y = public.public_numbers().y.to_bytes(32, "big")
        # kty: EC2, crv: P-256, x, y
        cosekey = {1: 2, -1: 1, -2: x, -3: y}
        if kid is not None:
            cosekey[2] = kid
        # cnf: COSE_Key
        credential_kccs: dict = {8: {1: cosekey}}
        if subject is not None:
            credential_kccs[2] = subject

        # kccs: cnf
        return {14: credential_kccs}


class EdhocCredentials(credentials._Objectish):
    own_key: Optional[CoseKeyForEdhoc]
    suite: int
    method: int
    own_cred: Optional[dict]
    peer_cred: Optional[dict]

    #: Whether the combined flow should be used when using this credential set.
    #:
    #: If unset or None, this the decision is left to the library (which at the
    #: time of writing always picks True).
    use_combined_edhoc: Optional[bool]

    def __init__(
        self,
        suite: int,
        method: int,
        own_cred_style: Optional[str] = None,
        peer_cred: Optional[dict] = None,
        own_cred: Optional[dict] = None,
        private_key_file: Optional[str] = None,
        private_key: Optional[dict] = None,
        use_combined_edhoc: Optional[bool] = None,
    ):
        from . import edhoc

        self.suite = suite
        self.method = method
        self.own_cred = own_cred
        self.peer_cred = peer_cred
        self.use_combined_edhoc = use_combined_edhoc

        if private_key_file is not None and private_key is not None:
            raise credentials.CredentialsLoadError(
                "private_key is mutually exclusive with private_key_file"
            )
        if private_key_file is not None:
            # FIXME: We should carry around a base
            private_key_path = Path(private_key_file)
            # FIXME: We left loading the file to the user, and now we're once more
            # in a position where we guess the file type
            self.own_key = CoseKeyForEdhoc.from_file(private_key_path)
        elif private_key is not None:
            self.own_key = CoseKeyForEdhoc.from_map(private_key)
        else:
            self.own_key = None

        if own_cred_style is None:
            self.own_cred_style = None
        else:
            self.own_cred_style = edhoc.OwnCredStyle(own_cred_style)

        if self.own_cred == {"unauthenticated": True}:
            if self.own_key is not None or self.own_cred_style not in (
                None,
                OwnCredStyle.ByValue,
            ):
                raise credentials.CredentialsLoadError(
                    "For local unauthenticated use, no key can be give, and own_cred_style needs to be by-value or absent"
                )
            self.own_key = CoseKeyForEdhoc.generate()
            # FIXME: kid and subj should rather not be sent, but lakers insists on their presence
            self.own_cred = self.own_key.as_ccs(kid=b"?", subject="")
            self.own_cred_style = OwnCredStyle.ByValue
        else:
            if (own_cred is None) != (own_cred_style is None) or (own_cred is None) != (
                self.own_key is None
            ):
                raise credentials.CredentialsLoadError(
                    "If own credentials are given, all of own_cred, own_cred_style and private_key(_path) need to be given"
                )

        # FIXME: This is only used on the client side, and expects that all parts (own and peer) are present
        self._established_context = None

    def find_edhoc_by_id_cred_peer(self, id_cred_peer):
        if self.peer_cred is None:
            return None
        if 14 not in self.peer_cred:
            # Only recognizing CCS so far
            return None

        if id_cred_peer == cbor2.dumps(self.peer_cred, canonical=True):
            # credential by value
            return cbor2.dumps(self.peer_cred[14], canonical=True)

        # cnf / COS_Key / kid, should be present in all CCS
        kid = self.peer_cred[14][8][1].get(2)
        if kid is not None and id_cred_peer == cbor2.dumps({4: kid}, canonical=True):
            # credential by kid
            return cbor2.dumps(self.peer_cred[14], canonical=True)

    def peer_cred_is_unauthenticated(self):
        # FIXME: This is rather weird internal API, and rather weird
        # format-wise -- but it will suffice until credentials are rewritten.
        return self.peer_cred is not None and self.peer_cred == {
            "unauthenticated": True
        }

    async def establish_context(
        self,
        wire,
        underlying_address,
        underlying_proxy_scheme,
        underlying_uri_host,
        logger,
    ):
        logger.info(
            "No OSCORE context found for EDHOC context %r, initiating one.", self
        )
        # The semantic identifier (an arbitrary string)
        #
        # FIXME: We don't support role reversal yet, but once we
        # register this context to be available for incoming
        # requests, we'll have to pick more carefully
        c_i = bytes([random.randint(0, 23)])
        initiator = lakers.EdhocInitiator()
        message_1 = initiator.prepare_message_1(c_i)

        msg1 = Message(
            code=POST,
            proxy_scheme=underlying_proxy_scheme,
            uri_host=underlying_uri_host,
            uri_path=[".well-known", "edhoc"],
            payload=cbor2.dumps(True) + message_1,
        )
        msg1.remote = underlying_address
        msg2 = await wire.request(msg1).response_raising

        (c_r, id_cred_r, ead_2) = initiator.parse_message_2(msg2.payload)

        assert isinstance(self.own_cred, dict) and list(self.own_cred.keys()) == [14], (
            "So far can only process CCS style own credentials a la {14: ...}, own_cred = %r"
            % self.own_cred
        )
        cred_i = cbor2.dumps(self.own_cred[14], canonical=True)
        key_i = self.own_key.d

        logger.debug("EDHOC responder sent message_2 with ID_CRED_R = %r", id_cred_r)
        if self.peer_cred_is_unauthenticated():
            # Not doing further checks (eg. for trailing bytes) or re-raising: This
            # was already checked by lakers
            parsed = cbor2.loads(id_cred_r)

            if 14 not in parsed:
                raise credentials.CredentialsMissingError(
                    "Peer presented credential-by-reference (or anything else that's not a KCCS) when no credential was pre-agreed"
                )

            # We could also pick the [14] out of the serialized stream and
            # treat it as an opaque item, but cbor2 doesn't really do serialized items.
            cred_r = cbor2.dumps(parsed[14], canonical=True)
        else:
            # We could look into id_cred_r, which is a CBOR encoded
            # byte string, and could start comparing ... but actually
            # EDHOC and Lakers protect us from misbinding attacks (is
            # that what they are called?), so we can just put in our
            # expected credential here
            #
            # FIXME: But looking into it might give us a better error than just
            # "Mac2 verification failed"

            # FIXME add assert on the structure or start doing the
            # generalization that'll fail at startup
            cred_r = cbor2.dumps(self.peer_cred[14], canonical=True)

        initiator.verify_message_2(
            key_i,
            cred_i,
            cred_r,
        )  # odd that we provide that here rather than in the next function

        logger.debug("Message 2 was verified")

        secctx = EdhocInitiatorContext(initiator, c_i, c_r, self.own_cred_style, logger)

        if self.use_combined_edhoc is not False:
            secctx.complete_without_message_4()
            # That's enough: Message 3 can well be sent along with the next
            # message.
            return secctx

        logger.debug("Sending explicit message 3 without optimization")

        message_3 = secctx._message_3
        assert message_3 is not None
        secctx._message_3 = None

        if len(c_r) == 1 and (0 <= c_r[0] < 24 or 0x20 <= c_r[0] < 0x38):
            c_r_encoded = c_r
        else:
            c_r_encoded = cbor2.dumps(c_r)
        msg3 = Message(
            code=POST,
            proxy_scheme=underlying_proxy_scheme,
            uri_host=underlying_uri_host,
            uri_path=[".well-known", "edhoc"],
            payload=c_r_encoded + message_3,
        )
        msg3.remote = underlying_address
        msg4 = await wire.request(msg3).response_raising

        secctx.complete_with_message_4(msg4.payload)

        logger.debug("Received message 4, context is ready")

        return secctx


class _EdhocContextBase(
    oscore.CanProtect, oscore.CanUnprotect, oscore.SecurityContextUtils
):
    def __init__(self, logger):
        self.log = logger

    def post_seqnoincrease(self):
        # The context is not persisted
        pass

    def protect(self, message, request_id=None, *, kid_context=True):
        outer_message, request_id = super().protect(
            message, request_id=request_id, kid_context=kid_context
        )
        message_3 = self.message_3_to_include()
        if message_3 is not None:
            outer_message.opt.edhoc = True
            outer_message.payload = message_3 + outer_message.payload
        return outer_message, request_id

    def _make_ready(self, edhoc_context, c_ours, c_theirs):
        # FIXME: both should offer this
        if (
            isinstance(edhoc_context, lakers.EdhocResponder)
            or edhoc_context.selected_cipher_suite() == 2
        ):
            self.alg_aead = oscore.algorithms["AES-CCM-16-64-128"]
            self.hashfun = oscore.hashfunctions["sha256"]
        else:
            raise RuntimeError("Unknown suite")

        # we did check for critical EADs, there was no out-of-band agreement, so 8 it is
        oscore_salt_length = 8
        # I figure that one would be ageed out-of-band as well (currently no
        # options to set/change this are known)
        self.id_context = None
        self.recipient_replay_window = oscore.ReplayWindow(32, lambda: None)

        master_secret = edhoc_context.edhoc_exporter(0, [], self.alg_aead.key_bytes)
        master_salt = edhoc_context.edhoc_exporter(1, [], oscore_salt_length)

        self.sender_id = c_theirs
        self.recipient_id = c_ours
        if self.sender_id == self.recipient_id:
            raise ValueError("Bad IDs: identical ones were picked")

        self.derive_keys(master_salt, master_secret)

        self.sender_sequence_number = 0
        self.recipient_replay_window.initialize_empty()

        self.log.debug("EDHOC context %r ready for OSCORE operation", self)

    @abc.abstractmethod
    def message_3_to_include(self) -> Optional[bytes]:
        """An encoded message_3 to include in outgoing messages

        This may modify self to only return something once."""


class EdhocInitiatorContext(_EdhocContextBase):
    """An OSCORE context that is derived from an EDHOC exchange.

    It does not require that the EDHOC exchange has completed -- it can be set
    up by an initiator already when message 2 has been received, prepares a
    message 3 at setup time, and sends it with the first request that is sent
    through it."""

    # FIXME: Should we rather send it with *every* request that is sent before a message 4 is received implicitly?
    def __init__(self, initiator, c_ours, c_theirs, cred_i_mode, logger):
        super().__init__(logger)

        # Only this line is role specific
        self._message_3, _i_prk_out = initiator.prepare_message_3(
            cred_i_mode.as_lakers(), None
        )

        self._incomplete = True
        self._init_details = (initiator, c_ours, c_theirs)

    def complete_without_message_4(self) -> None:
        (initiator, c_ours, c_theirs) = self._init_details
        initiator.completed_without_message_4()
        self._make_ready(initiator, c_ours, c_theirs)
        self._incomplete = False
        self._init_details = None

    def complete_with_message_4(self, message_4: bytes) -> None:
        (initiator, c_ours, c_theirs) = self._init_details
        initiator.process_message_4(message_4)
        self._make_ready(initiator, c_ours, c_theirs)
        self._incomplete = False
        self._init_details = None

    def message_3_to_include(self) -> Optional[bytes]:
        if self._message_3 is not None:
            result = self._message_3
            self._message_3 = None
            return result
        return None


class EdhocResponderContext(_EdhocContextBase):
    def __init__(self, responder, c_i, c_r, server_credentials, logger):
        super().__init__(logger)

        # storing them where they will later be overwritten with themselves
        self.recipient_id = c_r
        self.sender_id = c_i

        self._responder = responder
        # Through these we'll look up id_cred_i
        self._server_credentials = server_credentials

        self.authenticated_claims = []

        # Not sure why mypy even tolerates this -- we're clearly not ready for
        # a general protect/unprotect, and things only work because all
        # relevant functions get their checks introduced
        self._incomplete = True

        self._message_4 = None

    def message_3_to_include(self) -> Optional[bytes]:
        # as a responder we never send one
        return None

    def get_oscore_context_for(self, unprotected):
        if oscore.COSE_KID_CONTEXT in unprotected:
            return None
        if unprotected.get(oscore.COSE_KID) == self.recipient_id:
            return self

    def find_all_used_contextless_oscore_kid(self) -> set[bytes]:
        return set((self.recipient_id,))

    def protect(self, *args, **kwargs):
        if self._incomplete:
            raise RuntimeError(
                "EDHOC has not completed yet, waiting for message 3, can not protect own messages yet"
            )
        return super().protect(*args, **kwargs)

    def unprotect(self, protected_message, request_id=None):
        if self._incomplete:
            if not protected_message.opt.edhoc:
                self.log.error(
                    "OSCORE failed: No EDHOC message 3 received and none present"
                )
                raise error.BadRequest("EDHOC incomplete")

            payload_stream = io.BytesIO(protected_message.payload)
            # discarding result -- just need to have a point to split
            _ = cbor2.load(payload_stream)
            m3len = payload_stream.tell()
            message_3 = protected_message.payload[:m3len]

            self._offer_message_3(message_3)

            protected_message = protected_message.copy(
                edhoc=False, payload=protected_message.payload[m3len:]
            )

        return super().unprotect(protected_message, request_id)

    def _offer_message_3(self, message_3: bytes) -> bytes:
        """Send in a message 3, obtain a message 4

        It is safe to discard the output, eg. when the CoAP EDHOC option is
        used."""
        if self._incomplete:
            id_cred_i, ead_3 = self._responder.parse_message_3(message_3)
            if ead_3 is not None:
                self.log.error("Aborting EDHOC: EAD3 present")
                raise error.BadRequest

            try:
                (cred_i, claims) = self._server_credentials.find_edhoc_by_id_cred_peer(
                    id_cred_i
                )
            except KeyError:
                self.log.error(
                    "Aborting EDHOC: No credentials found for client with id_cred_i=h'%s'",
                    id_cred_i.hex(),
                )
                raise error.BadRequest

            self.authenticated_claims.extend(claims)

            self._responder.verify_message_3(cred_i)
            self._message_4 = self._responder.prepare_message_4()

            self._make_ready(self._responder, self.recipient_id, self.sender_id)
            self._incomplete = False

        return self._message_4


class OwnCredStyle(enum.Enum):
    """Guidance for how the own credential should be sent in an EDHOC
    exchange"""

    ByKeyId = "by-key-id"
    ByValue = "by-value"

    def as_lakers(self):
        """Convert the enum into Lakers' reepresentation of the same concept.

        The types may eventually be unified, but so far, Lakers doesn't make
        the distinctions we expect to make yet."""
        if self == self.ByKeyId:
            # FIXME: Mismatch to be fixed in lakers -- currently the only way
            # it allows sending by reference is by Key ID
            return lakers.CredentialTransfer.ByReference
        if self == self.ByValue:
            return lakers.CredentialTransfer.ByValue
        else:
            raise RuntimeError("enum variant not covered")



# SPDX-FileCopyrightText: Michael Burrows <mjb@asplake.co.uk>
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Parse and format link headers according to RFC 5988 "Web Linking".

Usage (assuming a suitable headers object in the environment):

>>> headers['Link'] = str(LinkHeader([Link("http://example.com/foo", rel="self"),
...                                   Link("http://example.com", rel="up")]))
>>> headers['Link']
'<http://example.com/foo>; rel=self, <http://example.com>; rel=up'
>>> parse(headers['Link'])
LinkHeader([Link('http://example.com/foo', rel='self'), Link('http://example.com', rel='up')])

Blank and missing values roundtrip correctly:

>>> format_link(parse('</s/1>; obs; if="core.s"; foo=""'))
'<</s/1>; obs; if=core.s; foo="">'

Conversions to and from json-friendly list-based structures are also provided:

>>> parse(headers['Link']).to_py()
[['http://example.com/foo', [['rel', 'self']]], ['http://example.com', [['rel', 'up']]]]
>>> str(LinkHeader([['http://example.com/foo', [['rel', 'self']]],
...                 ['http://example.com', [['rel', 'up']]]]))
'<http://example.com/foo>; rel=self, <http://example.com>; rel=up'

For further information see parse(), LinkHeader and Link.
"""

import re
from typing import Dict
from urllib.parse import urljoin

__all__ = [
    "parse",
    "format_links",
    "format_link",
    "LinkHeader",
    "Link",
    "ParseException",
]

SINGLE_VALUED_ATTRS = ["rel", "anchor", "rev", "media", "title", "title*", "type"]

#
# Regexes for link header parsing.  TOKEN and QUOTED in particular should conform to RFC2616.
#
# Acknowledgement: The QUOTED regexp is based on
# http://stackoverflow.com/questions/249791/regexp-for-quoted-string-with-escaping-quotes/249937#249937
#
# Trailing spaces are consumed by each pattern.  The RE_HREF pattern also allows for any leading spaces.
#

QUOTED = (
    r'"((?:[^"\\]|\\.)*)"'  # double-quoted strings with backslash-escaped double quotes
)
TOKEN = r"([^()<>@,;:\"\[\]?={}\s]+)"  # non-empty sequence of non-separator characters
RE_COMMA_HREF = re.compile(
    r" *,? *< *([^>]*) *> *"
)  # includes ',' separator; no attempt to check URI validity
RE_ONLY_TOKEN = re.compile(r"^%(TOKEN)s$" % locals())
RE_ATTR = re.compile(r"%(TOKEN)s *(?:= *(%(TOKEN)s|%(QUOTED)s))? *" % locals())
RE_SEMI = re.compile(r"; *")
RE_COMMA = re.compile(r", *")


def parse(header):
    """Parse a link header string, returning a LinkHeader object:

    >>> parse('<http://example.com/foo>; rel="foo bar", <http://example.com>; rel=up; type=text/html')
    LinkHeader([Link('http://example.com/foo', rel='foo bar'), Link('http://example.com', rel='up', type='text/html')])

    ParseException is raised in the event that the input string is not parsed completely:

    >>> parse('<http://example.com/foo> error') #doctest: +SKIP
    Traceback (most recent call last):
        ...
    ParseException: ('link_header.parse() failed near %s', "'error'")
    """
    scanner = _Scanner(header)
    links = []
    while scanner.scan(RE_COMMA_HREF):
        href = scanner[1]
        attrs = []
        while scanner.scan(RE_SEMI):
            if scanner.scan(RE_ATTR):
                attr_name, token, quoted = scanner[1], scanner[3], scanner[4]
                if quoted is not None:
                    attrs.append([attr_name, quoted.replace(r"\"", '"')])
                elif token is not None:
                    attrs.append([attr_name, token])
                else:
                    attrs.append([attr_name, None])
        links.append(Link(href, attrs))

    if scanner.buf:
        raise ParseException("link_header.parse() failed near %s", repr(scanner.buf))

    return LinkHeader(links)


def format_links(*args, **kwargs):
    return str(LinkHeader(*args, **kwargs))


def format_link(*args, **kwargs):
    return str(Link(*args, **kwargs))


class ParseException(Exception):
    pass


class LinkHeader(object):
    """Represents a sequence of links that can be formatted together as a link header."""

    def __init__(self, links=None):
        """Initializes a LinkHeader object with a list of Link objects or with
        list of parameters from which Link objects can be created:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])
        LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])
        >>> LinkHeader([['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]])
        LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])

        The Link objects can be accessed afterwards via the `links` property.

        String conversion follows the spec:

        >>> str(LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]))
        '<http://example.com/foo>; rel=foo, <http://example.com>; rel=up'

        Conversion to json-friendly list-based structures:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]).to_py()
        [['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]]

        """

        self.links = [
            link if isinstance(link, Link) else Link(*link) for link in links or []
        ]

    def to_py(self):
        """Supports list conversion:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]).to_py()
        [['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]]
        """
        return [link.to_py() for link in self.links]

    def __repr__(self):
        return "LinkHeader([%s])" % ", ".join(repr(link) for link in self.links)

    def __str__(self):
        """Formats a link header:

        >>> str(LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]))
        '<http://example.com/foo>; rel=foo, <http://example.com>; rel=up'
        """
        return ", ".join(str(link) for link in self.links)

    def links_by_attr_pairs(self, pairs):
        """Lists links that have attribute pairs matching all the supplied pairs:

        >>> parse('<http://example.com/foo>; rel="foo", <http://example.com>; rel="up"'
        ...      ).links_by_attr_pairs([('rel', 'up')])
        [Link('http://example.com', rel='up')]
        """
        return [
            link
            for link in self.links
            if all([key, value] in link.attr_pairs for key, value in pairs)
        ]


class Link(object):
    """Represents a single link."""

    def __init__(self, href, attr_pairs=None, **kwargs):
        """Initializes a Link object with an href and attributes either in
        the form of a sequence of key/value pairs &/or as keyword arguments.
        The sequence form allows to be repeated.  Attributes may be accessed
        subsequently via the `attr_pairs` property.

        String conversion follows the spec:

        >>> str(Link('http://example.com', [('foo', 'bar'), ('foo', 'baz')], rel='self'))
        '<http://example.com>; foo=bar; foo=baz; rel=self'

        Conversion to json-friendly list-based structures:

        >>> Link('http://example.com', [('foo', 'bar'), ('foo', 'baz')], rel='self').to_py()
        ['http://example.com', [['foo', 'bar'], ['foo', 'baz'], ['rel', 'self']]]
        """
        self.href = href
        self.attr_pairs = [
            list(pair) for pair in (attr_pairs or []) + list(kwargs.items())
        ]

    def to_py(self):
        """Convert to a json-friendly list-based structure:

        >>> Link('http://example.com', rel='foo').to_py()
        ['http://example.com', [['rel', 'foo']]]
        """
        return [self.href, self.attr_pairs]

    def __repr__(self):
        """
        >>> Link('http://example.com', rel='self')
        Link('http://example.com', rel='self')
        """
        return "Link(%s)" % ", ".join(
            [repr(self.href)]
            + ["%s=%s" % (pair[0], repr(pair[1])) for pair in self.attr_pairs]
        )

    def __str__(self):
        """Formats a single link:

        >>> str(Link('http://example.com/foo', [['rel', 'self']]))
        '<http://example.com/foo>; rel=self'
        >>> str(Link('http://example.com/foo', [['rel', '"quoted"'], ['type', 'text/html'], ['title*', "UTF-8'en'%e2%82%ac%20rates"]]))
        '<http://example.com/foo>; rel="\\\\"quoted\\\\""; type=text/html; title*=UTF-8\\'en\\'%e2%82%ac%20rates'

        Note that there is no explicit support for the title* attribute other
        than to output it unquoted.  Where used, it is up to client applications to
        provide values that meet RFC2231 Section 7.
        """

        def str_pair(key, value):
            if value is None:
                return key
            elif RE_ONLY_TOKEN.match(value) or key.endswith("*"):
                return "%s=%s" % (key, value)
            else:
                return '%s="%s"' % (key, value.replace('"', r"\""))

        return "; ".join(
            ["<%s>" % self.href]
            + [str_pair(key, value) for key, value in self.attr_pairs]
        )

    def __getattr__(self, name):
        """
        >>> Link('/', rel='self').rel
        'self'
        >>> Link('/', hreflang='EN').hreflang
        ['EN']
        >>> Link('/', foo='bar').foo
        ['bar']
        >>> Link('/', [('foo', 'bar'), ('foo', 'baz')]).foo
        ['bar', 'baz']
        >>> Link('/').rel #doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        AttributeError: No attribute of type 'rel' present
        >>> Link('/').hreflang
        []
        >>> Link('/').foo
        []
        """
        name_lower = name.lower()
        values = [value for key, value in self.attr_pairs if key.lower() == name_lower]
        if name in SINGLE_VALUED_ATTRS:
            if values:
                return values[0]
            else:
                raise AttributeError("No attribute of type %r present" % name_lower)
        return values

    def __contains__(self, name):
        """
        >>> 'rel' in Link('/', rel='self')
        True
        >>> 'obs' in Link('/', obs=None)
        True
        >>> 'rel' in Link('/')
        False
        """
        name_lower = name.lower()
        return any(key.lower() == name_lower for key, value in self.attr_pairs)

    def get_context(self, requested_resource_address):
        """Return the absolute URI of the context of a link. This is usually
        equals the base address the statement is about (eg. the requested URL
        if the link header was served in a successful HTTP GET request), but
        can be overridden by the anchor parameter.

        >>> Link('../', rel='index').get_context('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/chapter1/'
        >>> Link('', rel='next', anchor='../').get_context('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/'
        """
        if "anchor" in self:
            return urljoin(requested_resource_address, self.anchor)
        return requested_resource_address

    def get_target(self, requested_resource_address):
        """Return the absolute URI of the target of a link. It is determined by
        joining the address from which the link header was retrieved with the
        link-value (inside angular brackets) according to RFC3986 section 5.

        >>> Link('../', rel='index').get_target('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/'
        >>> Link('', rel='next', anchor='../').get_target('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/chapter1/'
        """
        return urljoin(requested_resource_address, self.href)


class _Scanner(object):
    def __init__(self, buf):
        self.buf = buf
        self.match = None

    def __getitem__(self, key):
        return self.match.group(key)

    def scan(self, pattern):
        self.match = pattern.match(self.buf)
        if self.match:
            self.buf = self.buf[self.match.end() :]
        return self.match


# For doctest
headers: Dict[str, str] = {}



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Client components for RFC9176_

Currently, this only contains the :mod:`register` module, which helps utilizing
the registration interface of a resource directory.

.. _RFC9176: https://datatracker.ietf.org/doc/html/rfc9176
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from gi.repository import Gtk
from gi.repository import Gdk

from aiocoap import resource
from aiocoap import Message, BAD_REQUEST

from .resource import *
from .util import _Throttler


class GobjectBacked(SenmlResource):
    """Provides a .value bound to the self.widget_property property of the
    self.backend_widget Gobject. The widget and widget_property need to be set
    before this class's __init__ is called."""

    def __init__(self):
        super().__init__()
        throttler = _Throttler(self.value_changed)
        self.backend_widget.connect(
            "notify::" + self.widget_property, lambda *k: throttler()
        )

    def _get_value(self):
        return self.backend_widget.get_property(self.widget_property)

    def _set_value(self, updated):
        self.backend_widget.set_property(self.widget_property, updated)

    value = property(_get_value, _set_value)


class Switch(BooleanResource, GobjectBacked):
    if_ = "core.s"
    # FIXME export as read-only

    widget_property = "state"

    def __init__(self, args):
        self.widget = self.backend_widget = Gtk.Switch()
        super().__init__()


class Bulb(BooleanResource):
    if_ = "core.a"
    # FIXME export as write-only

    def _get_value(self):
        # raise RuntimeError("This object is not readable") # this was a nice idea for a demo, but precludes even toggling
        return self.widget.props.icon_name == "weather-clear"

    def _set_value(self, new_state):
        self.widget.props.icon_name = {
            True: "weather-clear",
            False: "weather-clear-night",
        }[new_state]

    value = property(_get_value, _set_value)

    def __init__(self, args):
        super().__init__()
        self.widget = Gtk.Image()

    @ContenttypeRendered.empty_post_handler()
    def emptypost(self):
        self.value = not self.value


class RGBChannel(FloatResource, PythonBacked):
    if_ = "core.a"

    channel_name = property(lambda self: ("r", "g", "b")[self.channel])

    def __init__(self, channel):
        super().__init__()

        self.channel = channel
        self.value = 0  # FIXME this doesn't cater for notifications


class RGBRoot(SubsiteBatch):
    @ContenttypeRendered.get_handler("text/plain;charset=utf-8", default=True)
    def __regular_get(self):
        return "#" + "".join("%02x" % int(255 * c.value) for c in self.site.channels)

    @ContenttypeRendered.put_handler("text/plain;charset=utf-8", default=True)
    def render_put(self, payload):
        if len(payload) == 7 and payload[0:1] == b"#":
            values = tuple(
                int(payload[i : i + 2].decode("ascii"), 16) / 255 for i in (1, 3, 5)
            )
        else:
            return Message(code=BAD_REQUEST)

        # FIXME it'd be nice to update them in a way so that our own
        # value_changed only fires once
        for i, v in enumerate(values):
            self.site.channels[i].value = v


class RGBLight(resource.Site):
    def __init__(self, args):
        super().__init__()

        self.channels = [RGBChannel(i) for i in range(3)]

        for c in self.channels:
            self.add_resource([c.channel_name], c)
        rgbroot = RGBRoot(self)
        self.add_resource([], RGBRoot(self))

        rgbroot.add_valuechange_callback(self.trigger_repaint)

        self.widget = Gtk.DrawingArea()
        self.widget.set_size_request(200, 100)
        self.widget.connect("draw", self.cb_draw)

    def cb_draw(self, widget, cr):
        Gdk.cairo_set_source_rgba(
            cr,
            Gdk.RGBA(
                self.channels[0].value, self.channels[1].value, self.channels[2].value
            ),
        )
        cr.paint()

    def trigger_repaint(self):
        # this is assuming that GtkPaintable doesn't have its own window, which
        # seems not to have been the case for quite some time
        self.widget.queue_draw_area(
            0, 0, self.widget.get_allocated_width(), self.widget.get_allocated_height()
        )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This is a usage example of aiocoap that demonstrates how to implement a
simple server. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import datetime
import logging

import asyncio

import aiocoap.resource as resource
from aiocoap.numbers.contentformat import ContentFormat
import aiocoap


class Welcome(resource.Resource):
    representations = {
        ContentFormat.TEXT: b"Welcome to the demo server",
        ContentFormat.LINKFORMAT: b"</.well-known/core>,ct=40",
        # ad-hoc for application/xhtml+xml;charset=utf-8
        ContentFormat(65000): b'<html xmlns="http://www.w3.org/1999/xhtml">'
        b"<head><title>aiocoap demo</title></head>"
        b"<body><h1>Welcome to the aiocoap demo server!</h1>"
        b'<ul><li><a href="time">Current time</a></li>'
        b'<li><a href="whoami">Report my network address</a></li>'
        b"</ul></body></html>",
    }

    default_representation = ContentFormat.TEXT

    async def render_get(self, request):
        cf = (
            self.default_representation
            if request.opt.accept is None
            else request.opt.accept
        )
        try:
            return aiocoap.Message(payload=self.representations[cf], content_format=cf)
        except KeyError:
            raise aiocoap.error.UnsupportedContentFormat


class BlockResource(resource.Resource):
    """Example resource which supports the GET and PUT methods. It sends large
    responses, which trigger blockwise transfer."""

    def __init__(self):
        super().__init__()
        self.set_content(
            b"This is the resource's default content. It is padded "
            b"with numbers to be large enough to trigger blockwise "
            b"transfer.\n"
        )

    def set_content(self, content):
        self.content = content
        while len(self.content) <= 1024:
            self.content = self.content + b"0123456789\n"

    async def render_get(self, request):
        return aiocoap.Message(payload=self.content)

    async def render_put(self, request):
        print("PUT payload: %s" % request.payload)
        self.set_content(request.payload)
        return aiocoap.Message(code=aiocoap.CHANGED, payload=self.content)


class SeparateLargeResource(resource.Resource):
    """Example resource which supports the GET method. It uses asyncio.sleep to
    simulate a long-running operation, and thus forces the protocol to send
    empty ACK first."""

    def get_link_description(self):
        # Publish additional data in .well-known/core
        return dict(**super().get_link_description(), title="A large resource")

    async def render_get(self, request):
        await asyncio.sleep(3)

        payload = (
            "Three rings for the elven kings under the sky, seven rings "
            "for dwarven lords in their halls of stone, nine rings for "
            "mortal men doomed to die, one ring for the dark lord on his "
            "dark throne.".encode("ascii")
        )
        return aiocoap.Message(payload=payload)


class TimeResource(resource.ObservableResource):
    """Example resource that can be observed. The `notify` method keeps
    scheduling itself, and calles `update_state` to trigger sending
    notifications."""

    def __init__(self):
        super().__init__()

        self.handle = None

    def notify(self):
        self.updated_state()
        self.reschedule()

    def reschedule(self):
        self.handle = asyncio.get_event_loop().call_later(5, self.notify)

    def update_observation_count(self, count):
        if count and self.handle is None:
            print("Starting the clock")
            self.reschedule()
        if count == 0 and self.handle:
            print("Stopping the clock")
            self.handle.cancel()
            self.handle = None

    async def render_get(self, request):
        payload = datetime.datetime.now().strftime("%Y-%m-%d %H:%M").encode("ascii")
        return aiocoap.Message(payload=payload)


class WhoAmI(resource.Resource):
    async def render_get(self, request):
        text = ["Used protocol: %s." % request.remote.scheme]

        text.append("Request came from %s." % request.remote.hostinfo)
        text.append("The server address used %s." % request.remote.hostinfo_local)

        claims = list(request.remote.authenticated_claims)
        if claims:
            text.append(
                "Authenticated claims of the client: %s."
                % ", ".join(repr(c) for c in claims)
            )
        else:
            text.append("No claims authenticated.")

        return aiocoap.Message(content_format=0, payload="\n".join(text).encode("utf8"))


# logging setup

logging.basicConfig(level=logging.INFO)
logging.getLogger("coap-server").setLevel(logging.DEBUG)


async def main():
    # Resource tree creation
    root = resource.Site()

    root.add_resource(
        [".well-known", "core"], resource.WKCResource(root.get_resources_as_linkheader)
    )
    root.add_resource([], Welcome())
    root.add_resource(["time"], TimeResource())
    root.add_resource(["other", "block"], BlockResource())
    root.add_resource(["other", "separate"], SeparateLargeResource())
    root.add_resource(["whoami"], WhoAmI())

    await aiocoap.Context.create_server_context(root)

    # Run forever
    await asyncio.get_running_loop().create_future()


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This is a usage example of aiocoap that demonstrates how to implement a
simple client. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)


async def main():
    protocol = await Context.create_client_context()

    request = Message(code=GET, uri="coap://vs0.inf.ethz.ch/obs", observe=0)

    pr = protocol.request(request)

    r = await pr.response
    print("First response: %s\n%r" % (r, r.payload))

    async for r in pr.observation:
        print("Next result: %s\n%r" % (r, r.payload))

        pr.observation.cancel()
        break
    print("Loop ended, sticking around")
    await asyncio.sleep(50)


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import unittest

from . import common
from .test_server import (
    WithAsyncLoop,
    Destructing,
    WithClient,
    WithTestServer,
    CLEANUPTIME,
)
import aiocoap.proxy.client
import aiocoap.cli.proxy
from aiocoap.util import hostportjoin


class WithReverseProxy(WithAsyncLoop, Destructing):
    def setUp(self):
        super(WithReverseProxy, self).setUp()

        self.reverseproxy = aiocoap.cli.proxy.Main(
            [
                "--reverse",
                "--bind",
                hostportjoin(self.proxyhost, self.proxyport),
                "--namebased",
                "%s:%s" % (self.name_for_real_server, self.servernetloc),
                "--pathbased",
                "%s:%s" % ("/".join(self.path_for_real_server), self.servernetloc),
            ],
            loop=self.loop,
        )
        self.loop.run_until_complete(self.reverseproxy.initializing)

    def tearDown(self):
        super(WithReverseProxy, self).tearDown()
        self.loop.run_until_complete(self.reverseproxy.shutdown())

        self._del_to_be_sure("reverseproxy")

        self.loop.run_until_complete(asyncio.sleep(CLEANUPTIME))

    proxyport = 56839
    proxyhost = common.loopbackname_v6 or common.loopbackname_v46
    proxyaddress = "%s:%d" % (proxyhost, proxyport)

    name_for_real_server = "aliasedname"
    path_for_real_server = ("aliased", "name")


class TestReverseProxy(WithReverseProxy, WithClient, WithTestServer):
    @unittest.skipIf(
        common.using_simple6,
        "Some proxy tests fail with simple6 (https://github.com/chrysn/aiocoap/issues/88)",
    )
    def test_routing(self):
        yieldfrom = lambda f: self.loop.run_until_complete(f)

        def req():
            request = aiocoap.Message(code=aiocoap.GET)
            request.unresolved_remote = self.proxyaddress
            request.opt.uri_path = ("big",)
            return request

        request = req()

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code,
            aiocoap.BAD_REQUEST,
            "GET without hostname gave resource (something like BAD_REQUEST expected)",
        )

        request = req()
        request.opt.uri_host = self.name_for_real_server

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "GET with hostname based proxying was not successful)",
        )

        request = req()
        request.opt.uri_path = self.path_for_real_server + request.opt.uri_path

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "GET with path based proxying was not successful)",
        )

    @unittest.skipIf(
        common.using_simple6,
        "Some proxy tests fail with simple6 (https://github.com/chrysn/aiocoap/issues/88)",
    )
    def test_options(self):
        yieldfrom = lambda f: self.loop.run_until_complete(f)

        def req():
            request = aiocoap.Message(code=aiocoap.GET)
            request.unresolved_remote = self.proxyaddress
            request.opt.uri_path = ("big",)
            request.opt.uri_host = self.name_for_real_server
            return request

        request = req()

        request.opt.proxy_scheme = "coap"

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code,
            aiocoap.BAD_OPTION,
            "Reverse proxy supports proxying even though it shouldn't.",
        )

        request = req()
        request.opt.add_option(
            aiocoap.optiontypes.StringOption(2**10 + 2, "can't proxy this")
        )

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code, aiocoap.BAD_OPTION, "Proxy did not react to unsafe option."
        )

        request = req()
        request.opt.add_option(
            aiocoap.optiontypes.StringOption(2**10, "nothing to see here")
        )

        response = yieldfrom(self.client.request(request).response)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Proxy did not ignore to safe-to-forward option.",
        )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import shutil
from pathlib import Path

import cbor2 as cbor

from aiocoap import oscore

contextdir = Path(__file__).parent / "common-context"


class LoggingFilesystemSecurityContext(oscore.FilesystemSecurityContext):
    def _extract_external_aad(self, message, request_id, local_is_sender):
        result = super()._extract_external_aad(message, request_id, local_is_sender)
        print(
            "Verify: External AAD: bytes.fromhex(%r), %r"
            % (result.hex(), cbor.loads(result))
        )
        return result


class NotifyingPlugtestSecurityContext(oscore.FilesystemSecurityContext):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notification_hooks = []

    def notify(self):
        for x in self.notification_hooks:
            x()

    def post_seqnoincrease(self):
        super().post_seqnoincrease()
        self.notify()

    def _replay_window_changed(self):
        super()._replay_window_changed()
        self.notify()


class PlugtestFilesystemSecurityContext(
    LoggingFilesystemSecurityContext, NotifyingPlugtestSecurityContext
):
    pass


def get_security_context(
    contextname, contextcopy: Path, simulate_unclean_shutdown=False
):
    """Copy the base context (disambiguated by contextname in "ab", "cd") onto
    the path in contextcopy if it does not already exist, and load the
    resulting context with the given role. The context will be monkey-patched
    for debugging purposes.

    With the simulate_unclean_shutdown aprameter set to True, any existing
    replay window is removed from the loaded state."""
    if not contextcopy.exists():
        contextcopy.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree((contextdir / contextname), contextcopy)

        print("Context %s copied to %s" % (contextname, contextcopy))

    secctx = PlugtestFilesystemSecurityContext(contextcopy)

    if simulate_unclean_shutdown:
        secctx.recipient_replay_window._index = None
        secctx.recipient_replay_window._bitfield = None

    return secctx


def additional_verify(description, lhs, rhs):
    if lhs == rhs:
        print("Additional verify passed: %s" % description)
    else:
        print("Additional verify failed (%s != %s): %s" % (lhs, rhs, description))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tests for the aiocoap.credentials module. This does not test the
functionality of the ciphers, just the syntax of loading them."""

import unittest

from aiocoap import Message, GET
from aiocoap.credentials import CredentialsMap, DTLS
import aiocoap.defaults


class TestCredentialsLoad(unittest.TestCase):
    def test_load_empty(self):
        raw = {}
        m = CredentialsMap()
        m.load_from_dict(raw)
        self.assertEqual(type(m), CredentialsMap)
        self.assertEqual(len(m), 0)

    def test_dtls(self):
        raw = {
            "coaps://some-dtls-host/*": {
                "dtls": {
                    "psk": {"hex": "73-65-63-72-65-74-50-53-4b"},
                    "client-identity": b"Client_identity",
                }
            }
        }

        m = CredentialsMap()
        m.load_from_dict(raw)
        # note we can use the slash-free version here and still get the result
        # for //some-host/* due to the URI normalization rules
        message = Message(code=GET, uri="coaps://some-dtls-host")
        secmatch = m.credentials_from_request(message)
        self.assertEqual(type(secmatch), DTLS)
        self.assertEqual(secmatch.psk, b"secretPSK")
        self.assertEqual(secmatch.client_identity, b"Client_identity")

    @unittest.skipIf(
        aiocoap.defaults.oscore_missing_modules(),
        "Modules missing for loading OSCORE contexts: %s"
        % (aiocoap.defaults.oscore_missing_modules(),),
    )
    def test_oscore_filebased(self):
        from aiocoap.oscore import FilesystemSecurityContext

        raw = {
            "coap://some-oscore-host/*": {
                "oscore": {
                    "basedir": __file__.replace(
                        "test_credentials.py", "test_credentials_oscore_context/"
                    )
                }
            },
            "coaps://some-dtls-host/*": {
                "dtls": {
                    "psk": {"hex": "73-65-63-72-65-74-50-53-4b"},
                    "client-identity": b"Client_identity",
                }
            },
        }

        m = CredentialsMap()
        m.load_from_dict(raw)
        message = Message(code=GET, uri="coap://some-oscore-host/.well-known/core")
        secmatch = m.credentials_from_request(message)
        self.assertEqual(type(secmatch), FilesystemSecurityContext)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import unittest

from . import common
from .test_server import WithAsyncLoop, Destructing, WithClient, TestServer, CLEANUPTIME
from .test_client import TestClientWithSetHost
import aiocoap.proxy.client
import aiocoap.cli.proxy
from aiocoap.util import hostportjoin


class WithProxyServer(WithAsyncLoop, Destructing):
    def setUp(self):
        super(WithProxyServer, self).setUp()

        self.forwardproxy = aiocoap.cli.proxy.Main(
            ["--forward", "--bind", hostportjoin(self.proxyhost, self.proxyport)]
        )
        self.loop.run_until_complete(self.forwardproxy.initializing)

    def tearDown(self):
        super(WithProxyServer, self).tearDown()
        self.loop.run_until_complete(self.forwardproxy.shutdown())

        self._del_to_be_sure("forwardproxy")

        self.loop.run_until_complete(asyncio.sleep(CLEANUPTIME))

    proxyport = 56839
    proxyhost = common.loopbackname_v6 or common.loopbackname_v46
    proxyaddress = "%s:%d" % (proxyhost, proxyport)


class WithProxyClient(WithClient, WithProxyServer):
    def setUp(self):
        super(WithProxyClient, self).setUp()
        original_client_log = self.client.log
        self.client = aiocoap.proxy.client.ProxyForwarder(
            self.proxyaddress, self.client
        )
        self.client.log = original_client_log

    def tearDown(self):
        self.client = self.client.context


class TestServerWithProxy(WithProxyClient, TestServer):
    def build_request(self):
        # this needs to be run differently because tests/server.py
        # doesn't exactly use the high-level apis. (and that's ok because we need
        # to test the server with simple messages too.)

        request = aiocoap.Message(code=aiocoap.GET)
        request.unresolved_remote = self.proxyaddress
        request.opt.proxy_scheme = "coap"
        request.opt.uri_host = self.serveraddress
        return request

    test_replacing_resource = unittest.skipIf(
        common.using_simple6,
        "Some proxy tests fail with simple6 (https://github.com/chrysn/aiocoap/issues/88)",
    )(TestServer.test_replacing_resource)
    test_slowbig_resource = unittest.skipIf(
        common.using_simple6,
        "Some proxy tests fail with simple6 (https://github.com/chrysn/aiocoap/issues/88)",
    )(TestServer.test_slowbig_resource)


# leaving that out for a moment because it fails more slowly

# class TestClientWithProxy(WithProxyClient, TestClientWithSetHost):
#    pass

# no need to run them again
del TestClientWithSetHost
del TestServer

# none of those tests would currently work, disabling them all. see
# https://github.com/chrysn/aiocoap/issues/106
del TestServerWithProxy



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This file is intended to be run in pyodide by the accompanying index.html"""

import aiocoap
import aiocoap.proxy.client
import js
import asyncio
import html


async def outer_main():
    status = js.document.getElementById("status").firstChild
    status.data = "Starting up"

    try:
        await main()
    except BaseException as e:
        status.data = f"Crashed: {e!r}"
        raise
    else:
        status.data = "Terminated (?)"


async def main():
    output = js.document.getElementById("output")
    status = js.document.getElementById("status").firstChild

    uri = js.document.getElementById("uri")
    proxy = js.document.getElementById("proxy")
    go = js.document.getElementById("go")
    mainform = js.document.getElementById("mainform")

    clicks = asyncio.Queue()

    # the proxy URI is absolute and even pathless, so no fragment should be in there anyway
    fragment = js.window.location.hash.removeprefix("#")
    if fragment:
        (proxy.value, _, uri.value) = fragment.partition("#")

    def onsubmit(event):
        clicks.put_nowait(None)
        return False

    mainform.onsubmit = onsubmit
    uri.disabled = False
    proxy.disabled = False
    go.disabled = False

    protocol = await aiocoap.Context.create_client_context()

    while True:
        output.innerHTML = ""
        js.document.title = uri.value
        js.window.location.hash = "#" + proxy.value + "#" + uri.value
        task = asyncio.create_task(
            observe(protocol, status, output, uri.value, proxy.value)
        )
        await clicks.get()
        task.cancel()
        status.value = "Starting over with new address…"


async def observe(protocol, status, output, uri, proxy):
    # FIXME: udp6 doesn't fail prettily on pyodide, catching
    if uri.partition("://")[0] not in ["coap+ws", "coaps+ws"] and not proxy:
        status.data = "Error: Only the URI schemes 'coap+ws' or 'coaps+ws' are supported without a configured proxy."
        return
    try:
        status.data = "Sending request…"
        while True:
            msg = aiocoap.Message(code=aiocoap.GET, uri=uri, observe=0)

            if proxy:
                # FIXME: aiocoap's proxying does not set proxy-scheme right, doing it manually
                # requester = aiocoap.proxy.client.ProxyForwarder(proxy, protocol)
                requester = protocol

                host, port = aiocoap.util.hostportsplit(msg.remote.hostinfo)
                msg.opt.uri_port = port
                msg.opt.uri_host = host
                msg.opt.proxy_scheme = msg.remote.scheme

                to_proxy = aiocoap.Message(uri=proxy)
                msg.remote = to_proxy.remote

                # and furthermore there is something wrong with the proxy
                # (sending an observable result and then 5.00), so skipping
                # observe here
                msg.opt.observe = None
            else:
                requester = protocol

            request = requester.request(msg)

            def render(response):
                status.data = f"Live data ({response.code})…"
                prettified = pretty(response)
                output.innerHTML = (
                    prettified
                    or '<p style="color:gray;font-size:small;">Response was empty</p>'
                )

            first = await request.response
            render(first)

            if first.opt.observe is None:
                status.data = f"Received response ({first.code})"
                break

            async for response in request.observation:
                render(response)

            status.data += ", and then lost observation; waiting before retry…"
            await asyncio.sleep(10)
            status.data = "Retrying…"
    except Exception as e:
        status.data = f"Failed permanently, please try different URI: {e!r}"


def pretty(message):
    # FIXME: This is part of the Message._repr_html_, but should be available
    # for standalone use without rendering code and options
    from aiocoap.util.prettyprint import pretty_print, lexer_for_mime

    (notes, mediatype, text) = pretty_print(message)
    import pygments
    from pygments.formatters import HtmlFormatter

    try:
        lexer = lexer_for_mime(mediatype)
        text = pygments.highlight(text, lexer, HtmlFormatter())
    except pygments.util.ClassNotFound:
        text = html.escape(text)
    return (
        "<div>"
        + "".join(
            f'<p style="color:gray;font-size:small;">{html.escape(n)}</p>'
            for n in notes
        )
        + f"<pre>{text}</pre>"
        + "</div>"
    )
